import asyncio
import logging
import os
from copy import deepcopy
from datetime import datetime
from typing import Callable, Dict, List, Optional

import yaml

from khoj.database.adapters import AgentAdapters, EntryAdapters, McpServerAdapters
from khoj.database.models import Agent, ChatMessageModel, KhojUser, UserMemory
from khoj.processor.conversation import prompts
from khoj.processor.conversation.utils import (
    OperatorRun,
    ResearchIteration,
    ToolCall,
    construct_iteration_history,
    construct_tool_chat_history,
    load_complex_json,
)
from khoj.processor.operator import operate_environment
from khoj.processor.tools.mcp import MCPClient
from khoj.processor.tools.online_search import read_webpages_content, search_online
from khoj.processor.tools.run_code import run_code
from khoj.routers.helpers import (
    ChatEvent,
    get_message_from_queue,
    grep_files,
    list_files,
    search_documents,
    send_message_to_model_wrapper,
    view_file_content,
)
from khoj.utils.helpers import (
    ConversationCommand,
    ToolDefinition,
    dict_to_tuple,
    is_code_sandbox_enabled,
    is_none_or_empty,
    is_operator_enabled,
    is_web_search_enabled,
    timer,
    tools_for_research_llm,
    truncate_code_context,
)
from khoj.utils.rawconfig import LocationData

logger = logging.getLogger(__name__)


class ToolExecutionResult:
    """Result of executing a single tool call"""

    def __init__(self):
        self.status_messages: List = []
        self.document_results: List[Dict[str, str]] = []
        self.online_results: Dict = {}
        self.code_results: Dict = {}
        self.operator_results: OperatorRun = None
        self.mcp_results: List = []
        self.should_terminate: bool = False


async def execute_tool(
    iteration: ResearchIteration,
    user: KhojUser,
    conversation_id: str,
    previous_iterations: List[ResearchIteration],
    location: LocationData,
    query_images: List[str],
    query_files: str,
    max_document_searches: int,
    max_online_searches: int,
    mcp_clients: List[MCPClient],
    cancellation_event: Optional[asyncio.Event],
    interrupt_queue: Optional[asyncio.Queue],
    agent: Agent,
    tracer: dict,
) -> ToolExecutionResult:
    """Execute a single tool call and return results. Designed for parallel execution."""
    result = ToolExecutionResult()

    # Skip if warning present
    if iteration.warning:
        logger.warning(f"Research mode: {iteration.warning}.")
        return result

    # Check for termination conditions
    if not iteration.query or isinstance(iteration.query, str) or iteration.query.name == ConversationCommand.Text:
        result.should_terminate = True
        return result

    # Create a status collector that captures messages for later batch yield
    async def status_collector(message: str):
        """Async generator that collects status messages instead of streaming them."""
        # Just collect the message - we'll process it later in the main loop
        result.status_messages.append(message)
        yield message  # Yield to satisfy async generator protocol expected by tool functions

    try:
        if iteration.query.name == ConversationCommand.SemanticSearchFiles:
            iteration.context = []
            previous_inferred_queries = {
                c["query"] for iter in previous_iterations if iter.context for c in iter.context
            }
            async for res in search_documents(
                **iteration.query.args,
                n=max_document_searches,
                d=None,
                user=user,
                chat_history=construct_tool_chat_history(previous_iterations, ConversationCommand.SemanticSearchFiles),
                conversation_id=conversation_id,
                conversation_commands=[ConversationCommand.Notes],
                location_data=location,
                send_status_func=status_collector,
                query_images=query_images,
                query_files=query_files,
                previous_inferred_queries=previous_inferred_queries,
                agent=agent,
                tracer=tracer,
            ):
                # Status messages are collected by status_collector, skip ChatEvent.STATUS here
                if isinstance(res, tuple):
                    result.document_results = res[0]
                    iteration.context += result.document_results

            if not is_none_or_empty(result.document_results):
                try:
                    distinct_files = {d["file"] for d in result.document_results}
                    distinct_headings = set(
                        [d["compiled"].split("\n")[0] for d in result.document_results if "compiled" in d]
                    )
                    headings_str = "\n- " + "\n- ".join(distinct_headings).replace("#", "")
                    async for _ in status_collector(
                        f"**Found {len(distinct_headings)} Notes Across {len(distinct_files)} Files**: {headings_str}"
                    ):
                        pass
                except Exception as e:
                    iteration.warning = f"Error extracting document references: {e}"
                    logger.error(iteration.warning, exc_info=True)
            else:
                iteration.warning = "No matching document references found"

        elif iteration.query.name == ConversationCommand.SearchWeb:
            previous_subqueries = {
                subquery for iter in previous_iterations if iter.onlineContext for subquery in iter.onlineContext.keys()
            }
            async for res in search_online(
                **iteration.query.args,
                conversation_history=construct_tool_chat_history(previous_iterations, ConversationCommand.SearchWeb),
                location=location,
                user=user,
                send_status_func=status_collector,
                custom_filters=[],
                max_online_searches=max_online_searches,
                max_webpages_to_read=0,
                query_images=query_images,
                previous_subqueries=previous_subqueries,
                agent=agent,
                tracer=tracer,
            ):
                # Status messages are collected by status_collector, skip ChatEvent.STATUS here
                if is_none_or_empty(res):
                    iteration.warning = (
                        "Detected previously run online search queries. Skipping iteration. Try something different."
                    )
                elif not (isinstance(res, dict) and ChatEvent.STATUS in res):
                    result.online_results = res
                    iteration.onlineContext = result.online_results

        elif iteration.query.name == ConversationCommand.ReadWebpage:
            async for res in read_webpages_content(
                **iteration.query.args,
                user=user,
                send_status_func=status_collector,
                agent=agent,
                tracer=tracer,
            ):
                # Status messages are collected by status_collector, skip ChatEvent.STATUS here
                if not (isinstance(res, dict) and ChatEvent.STATUS in res):
                    direct_web_pages: Dict[str, Dict] = res
                    for web_query in direct_web_pages:
                        if result.online_results.get(web_query):
                            result.online_results[web_query]["webpages"] = direct_web_pages[web_query]["webpages"]
                        else:
                            result.online_results[web_query] = {"webpages": direct_web_pages[web_query]["webpages"]}
                    iteration.onlineContext = result.online_results

        elif iteration.query.name == ConversationCommand.PythonCoder:
            async for res in run_code(
                **iteration.query.args,
                conversation_history=construct_tool_chat_history(previous_iterations, ConversationCommand.PythonCoder),
                context="",
                location_data=location,
                user=user,
                send_status_func=status_collector,
                query_images=query_images,
                query_files=query_files,
                agent=agent,
                tracer=tracer,
            ):
                # Status messages are collected by status_collector, skip ChatEvent.STATUS here
                if not (isinstance(res, dict) and ChatEvent.STATUS in res):
                    result.code_results = res
                    iteration.codeContext = result.code_results
            if iteration.codeContext:
                async for _ in status_collector(f"**Ran code snippets**: {len(iteration.codeContext)}"):
                    pass

        elif iteration.query.name == ConversationCommand.ViewFile:
            async for res in view_file_content(
                **iteration.query.args,
                user=user,
            ):
                if res and isinstance(res, list):
                    if iteration.context is None:
                        iteration.context = []
                    result.document_results = res
                    iteration.context += result.document_results
            async for _ in status_collector(f"**Viewed file**: {iteration.query.args['path']}"):
                pass

        elif iteration.query.name == ConversationCommand.ListFiles:
            async for res in list_files(
                **iteration.query.args,
                user=user,
            ):
                # Status messages are collected by status_collector, skip ChatEvent.STATUS here
                if not (isinstance(res, dict) and ChatEvent.STATUS in res):
                    if iteration.context is None:
                        iteration.context = []
                    result.document_results = [res]
                    iteration.context += result.document_results
            if result.document_results:
                async for _ in status_collector(result.document_results[-1].get("query", "Listed files")):
                    pass

        elif iteration.query.name == ConversationCommand.RegexSearchFiles:
            async for res in grep_files(
                **iteration.query.args,
                user=user,
            ):
                # Status messages are collected by status_collector, skip ChatEvent.STATUS here
                if not (isinstance(res, dict) and ChatEvent.STATUS in res):
                    if iteration.context is None:
                        iteration.context = []
                    result.document_results = [res]
                    iteration.context += result.document_results
            if result.document_results:
                async for _ in status_collector(result.document_results[-1].get("query", "Searched files")):
                    pass

        elif "/" in iteration.query.name:
            server_name, tool_name = iteration.query.name.split("/", 1)
            mcp_client = next((client for client in mcp_clients if client.name == server_name), None)
            if not mcp_client:
                raise ValueError(f"Could not find MCP server with name {server_name}")

            result.mcp_results = await mcp_client.run_tool(tool_name, iteration.query.args)
            if iteration.context is None:
                iteration.context = []
            iteration.context += result.mcp_results
            async for _ in status_collector(f"**Used MCP Tool**: {tool_name} on {mcp_client.name}"):
                pass

        else:
            result.should_terminate = True

    except Exception as e:
        tool_name = iteration.query.name if iteration.query else "unknown"
        iteration.warning = f"Error executing {tool_name}: {e}"
        logger.error(iteration.warning, exc_info=True)

    return result


async def apick_next_tool(
    query: str,
    conversation_history: List[ChatMessageModel],
    user: KhojUser = None,
    location: LocationData = None,
    user_name: str = None,
    agent: Agent = None,
    previous_iterations: List[ResearchIteration] = [],
    max_iterations: int = 5,
    query_images: List[str] = [],
    query_files: str = None,
    relevant_memories: List[UserMemory] = [],
    max_document_searches: int = 7,
    max_online_searches: int = 3,
    max_webpages_to_read: int = 3,
    mcp_clients: List[MCPClient] = [],
    send_status_func: Optional[Callable] = None,
    tracer: dict = {},
):
    """Given a query, determine which of the available tools the agent should use in order to answer appropriately."""

    # Continue with previous iteration if a multi-step tool use is in progress
    if (
        previous_iterations
        and previous_iterations[-1].query
        and isinstance(previous_iterations[-1].query, ToolCall)
        and previous_iterations[-1].query.name == ConversationCommand.Operator
        and not previous_iterations[-1].summarizedResult
    ):
        previous_iteration = previous_iterations[-1]
        yield ResearchIteration(
            query=ToolCall(name=previous_iteration.query.name, args={"query": query}, id=previous_iteration.query.id),  # type: ignore
            context=previous_iteration.context,
            onlineContext=previous_iteration.onlineContext,
            codeContext=previous_iteration.codeContext,
            operatorContext=previous_iteration.operatorContext,
            warning=previous_iteration.warning,
        )
        return

    # Construct tool options for the agent to choose from
    tools = []
    tool_options_str = ""
    agent_input_tools = agent.input_tools if agent and agent.input_tools else []
    agent_tools = []

    # Map agent user facing tools to research tools to include in agents toolbox
    document_research_tools = [
        ConversationCommand.SemanticSearchFiles,
        ConversationCommand.RegexSearchFiles,
        ConversationCommand.ViewFile,
        ConversationCommand.ListFiles,
    ]
    web_research_tools = [ConversationCommand.SearchWeb, ConversationCommand.ReadWebpage]
    input_tools_to_research_tools = {
        ConversationCommand.Notes.value: [tool.value for tool in document_research_tools],
        ConversationCommand.Webpage.value: [ConversationCommand.ReadWebpage.value],
        ConversationCommand.Online.value: [ConversationCommand.SearchWeb.value],
        ConversationCommand.Code.value: [ConversationCommand.PythonCoder.value],
        ConversationCommand.Operator.value: [ConversationCommand.OperateComputer.value],
    }
    for input_tool, research_tools in input_tools_to_research_tools.items():
        if input_tool in agent_input_tools:
            agent_tools += research_tools

    user_has_entries = await EntryAdapters.auser_has_entries(user)
    for tool, tool_data in tools_for_research_llm.items():
        # Skip showing operator tool as an option if not enabled
        if tool == ConversationCommand.OperateComputer and not is_operator_enabled():
            continue
        # Skip showing document related tools if user has no documents
        if tool in document_research_tools and not user_has_entries:
            continue
        # Skip showing web search tool if agent has no access to internet
        if tool in web_research_tools and not is_web_search_enabled():
            continue
        # Skip showing code tool if agent has no access to code execution sandbox
        if tool == ConversationCommand.PythonCoder and not is_code_sandbox_enabled():
            continue
        # Format description with relevant usage limits
        if tool == ConversationCommand.SemanticSearchFiles:
            description = tool_data.description.format(max_search_queries=max_document_searches)
        elif tool == ConversationCommand.ReadWebpage:
            description = tool_data.description.format(max_webpages_to_read=max_webpages_to_read)
        elif tool == ConversationCommand.SearchWeb:
            description = tool_data.description.format(max_search_queries=max_online_searches)
        else:
            description = tool_data.description
        # Add tool if agent does not have any tools defined or the tool is supported by the agent.
        if len(agent_tools) == 0 or tool.value in agent_tools:
            tool_options_str += f'- "{tool.value}": "{description}"\n'
            tools.append(
                ToolDefinition(
                    name=tool.value,
                    description=description,
                    schema=tool_data.schema,
                )
            )

    # Get MCP tools
    for mcp_client in mcp_clients:
        try:
            mcp_tools = await mcp_client.get_tools()
            for mcp_tool in mcp_tools:
                qualified_tool_name = f"{mcp_client.name}/{mcp_tool.name}"
                tool_options_str += f'- "{qualified_tool_name}": "{mcp_tool.description}"\n'
                tools.append(
                    ToolDefinition(
                        name=qualified_tool_name,
                        description=mcp_tool.description,
                        schema=mcp_tool.inputSchema,
                    )
                )
        except Exception as e:
            logger.warning(f'Failed to get tools from MCP server "{mcp_client.name}", so skipping: {e}.')

    today = datetime.today()
    location_data = f"{location}" if location else "Unknown"
    agent_chat_model = AgentAdapters.get_agent_chat_model(agent, user) if agent else None
    personality_context = (
        prompts.personality_context.format(personality=agent.personality) if agent and agent.personality else ""
    )

    function_planning_prompt = prompts.plan_function_execution.format(
        tools=tool_options_str,
        personality_context=personality_context,
        current_date=today.strftime("%Y-%m-%d"),
        day_of_week=today.strftime("%A"),
        username=user_name or "Unknown",
        location=location_data,
        max_iterations=max_iterations,
    )

    # Construct chat history with user and iteration history with researcher agent for context
    iteration_chat_history = construct_iteration_history(previous_iterations, query, query_images, query_files)
    chat_and_research_history = conversation_history + iteration_chat_history

    try:
        with timer("Chat actor: Infer information sources to refer", logger):
            response = await send_message_to_model_wrapper(
                query="",
                query_files=query_files,
                query_images=query_images,
                relevant_memories=relevant_memories,
                system_message=function_planning_prompt,
                chat_history=chat_and_research_history,
                tools=tools,
                deepthought=True,
                fast_model=False,
                agent_chat_model=agent_chat_model,
                user=user,
                tracer=tracer,
            )
    except Exception as e:
        logger.error(f"Failed to infer information sources to refer: {e}", exc_info=True)
        yield ResearchIteration(
            query=None,
            warning="Failed to infer information sources to refer. Skipping iteration. Try again.",
        )
        return

    try:
        # Try parse the response as function call response to infer next tools to use.
        response_text = response.text
        parsed_responses = [ToolCall(**item) for item in load_complex_json(response_text)]
    except Exception:
        # Otherwise assume the model has decided to end the research run and respond to the user.
        parsed_responses = [ToolCall(name=ConversationCommand.Text, args={"response": response_text}, id=None)]

    # Detect selection of previously used query, tool combinations.
    previous_tool_query_combinations = {
        (i.query.name, dict_to_tuple(i.query.args))
        for i in previous_iterations
        if i.warning is None and isinstance(i.query, ToolCall)
    }

    # Send status update with model's thoughts if available
    if send_status_func and not is_none_or_empty(response.thought):
        async for event in send_status_func(response.thought):
            yield {ChatEvent.STATUS: event}

    # Yield a ResearchIteration for each tool call to enable parallel execution
    for idx, parsed_response in enumerate(parsed_responses):
        warning = None
        logger.info(
            f"Response for determining relevant tools ({idx + 1}/{len(parsed_responses)}): {parsed_response.name}({parsed_response.args})"
        )

        if (parsed_response.name, dict_to_tuple(parsed_response.args)) in previous_tool_query_combinations:
            warning = f"Repeated tool, query combination detected. You've already called {parsed_response.name} with args: {parsed_response.args}. Try something different."

        # Include raw_response only for the first tool call to avoid duplication in history
        yield ResearchIteration(
            query=parsed_response,
            warning=warning,
            raw_response=response.raw_content if idx == 0 else None,
        )


async def research(
    user: KhojUser,
    query: str,
    conversation_id: str,
    conversation_history: List[ChatMessageModel],
    previous_iterations: List[ResearchIteration],
    query_images: List[str],
    query_files: str = None,
    relevant_memories: List[UserMemory] = [],
    user_name: str = None,
    location: LocationData = None,
    send_status_func: Optional[Callable] = None,
    cancellation_event: Optional[asyncio.Event] = None,
    interrupt_queue: Optional[asyncio.Queue] = None,
    abort_message: str = ChatEvent.END_EVENT.value,
    agent: Agent = None,
    tracer: dict = {},
):
    max_document_searches = 7
    max_online_searches = 3
    max_webpages_to_read = 1
    current_iteration = 0
    MAX_ITERATIONS = int(os.getenv("KHOJ_RESEARCH_ITERATIONS", 5))

    # Construct MCP clients
    mcp_servers = await McpServerAdapters.aget_all_mcp_servers()
    mcp_clients = [MCPClient(server.name, server.path, server.api_key) for server in mcp_servers]

    # Incorporate previous partial research into current research chat history
    research_conversation_history = [chat for chat in deepcopy(conversation_history) if chat.message]
    if (current_iteration := len(previous_iterations)) > 0:
        logger.info(f"Continuing research with the previous {len(previous_iterations)} iteration results.")
        previous_iterations_history = construct_iteration_history(previous_iterations)
        research_conversation_history += previous_iterations_history

    while current_iteration < MAX_ITERATIONS:
        # Check for cancellation at the start of each iteration
        if cancellation_event and cancellation_event.is_set():
            logger.debug(f"Research cancelled. User {user} disconnected client.")
            break

        # Update the query for the current research iteration
        if interrupt_query := get_message_from_queue(interrupt_queue):
            if interrupt_query == abort_message:
                cancellation_event.set()
                logger.debug(f"Research cancelled by user {user} via interrupt queue.")
                break
            # Add the interrupt query as a new user message to the research conversation history
            logger.info(
                f"Continuing research for user {user} with the previous {len(previous_iterations)} iterations and new instruction: {interrupt_query}"
            )
            previous_iterations_history = construct_iteration_history(
                previous_iterations, query, query_images, query_files
            )
            research_conversation_history += previous_iterations_history
            query = interrupt_query
            previous_iterations = []

            async for result in send_status_func(f"**Incorporate New Instruction**: {interrupt_query}"):
                yield result

        # Collect all tool calls from apick_next_tool
        iterations_to_process: List[ResearchIteration] = []
        async for result in apick_next_tool(
            query,
            research_conversation_history,
            user,
            location,
            user_name,
            agent,
            previous_iterations,
            MAX_ITERATIONS,
            query_images=query_images,
            query_files=query_files,
            relevant_memories=relevant_memories,
            max_document_searches=max_document_searches,
            max_online_searches=max_online_searches,
            max_webpages_to_read=max_webpages_to_read,
            mcp_clients=mcp_clients,
            send_status_func=send_status_func,
            tracer=tracer,
        ):
            if isinstance(result, dict) and ChatEvent.STATUS in result:
                yield result[ChatEvent.STATUS]
            elif isinstance(result, ResearchIteration):
                iterations_to_process.append(result)
                yield result

        # Multi-turn tools that stream their execution
        streaming_tools = {ConversationCommand.OperateComputer}
        if iterations_to_process:
            # Separate streaming tools that need real-time status updates
            # from parallelizable tools that can batch their status messages
            streaming_iterations: list[ResearchIteration] = []
            parallel_iterations: list[ResearchIteration] = []
            for iteration in iterations_to_process:
                if isinstance(iteration.query, ToolCall) and iteration.query.name in streaming_tools:
                    streaming_iterations.append(iteration)
                else:
                    parallel_iterations.append(iteration)

            # Execute streaming tools sequentially for real-time status updates
            streaming_results: list[tuple[ResearchIteration, ToolExecutionResult]] = []
            for iteration in streaming_iterations:
                result = ToolExecutionResult()
                if (
                    isinstance(iteration.query, ToolCall)
                    and iteration.query.name == ConversationCommand.OperateComputer
                ):
                    try:
                        # Execute OperateComputer
                        async for res in operate_environment(
                            **iteration.query.args,
                            user=user,
                            conversation_log=construct_tool_chat_history(
                                previous_iterations, ConversationCommand.Operator
                            ),
                            location_data=location,
                            previous_trajectory=previous_iterations[-1].operatorContext
                            if previous_iterations
                            else None,
                            send_status_func=send_status_func,
                            query_images=query_images,
                            agent=agent,
                            query_files=query_files,
                            cancellation_event=cancellation_event,
                            interrupt_queue=interrupt_queue,
                            tracer=tracer,
                        ):
                            if isinstance(res, dict) and ChatEvent.STATUS in res:
                                yield res[ChatEvent.STATUS]
                            elif isinstance(res, OperatorRun):
                                result.operator_results = res
                                iteration.operatorContext = result.operator_results
                                if res.webpages:
                                    if not result.online_results.get(iteration.query):
                                        result.online_results[iteration.query] = {"webpages": res.webpages}
                                    elif not result.online_results[iteration.query].get("webpages"):
                                        result.online_results[iteration.query]["webpages"] = res.webpages
                                    else:
                                        result.online_results[iteration.query]["webpages"] += res.webpages
                                    iteration.onlineContext = result.online_results
                    except Exception as e:
                        iteration.warning = f"Error operating browser: {e}"
                        logger.error(iteration.warning, exc_info=True)
                streaming_results.append((iteration, result))

            # Execute parallelizable tools in parallel
            parallel_results = []
            if parallel_iterations:
                tasks = [
                    execute_tool(
                        iteration=iteration,
                        user=user,
                        conversation_id=conversation_id,
                        previous_iterations=previous_iterations,
                        location=location,
                        query_images=query_images,
                        query_files=query_files,
                        max_document_searches=max_document_searches,
                        max_online_searches=max_online_searches,
                        mcp_clients=mcp_clients,
                        cancellation_event=cancellation_event,
                        interrupt_queue=interrupt_queue,
                        agent=agent,
                        tracer=tracer,
                    )
                    for iteration in parallel_iterations
                ]
                tool_results = await asyncio.gather(*tasks, return_exceptions=True)
                parallel_results = list(zip(parallel_iterations, tool_results))

            # Combine results (streaming first, then parallel)
            all_results = streaming_results + parallel_results

            # Process results and yield status messages
            for this_iteration, tool_result in all_results:
                # Handle exceptions from asyncio.gather
                if isinstance(tool_result, Exception):
                    this_iteration.warning = f"Error executing tool: {tool_result}"
                    logger.error(this_iteration.warning, exc_info=True)
                    tool_result = ToolExecutionResult()

                # Check for termination
                if tool_result.should_terminate:
                    current_iteration = MAX_ITERATIONS

                # Yield all collected status messages through the real send_status_func
                for status_msg in tool_result.status_messages:
                    if send_status_func:
                        async for status_event in send_status_func(status_msg):
                            yield status_event

                current_iteration += 1

                # Build summarized results
                if (
                    tool_result.document_results
                    or tool_result.online_results
                    or tool_result.code_results
                    or tool_result.operator_results
                    or tool_result.mcp_results
                    or this_iteration.warning
                ):
                    results_data = f"\n<iteration_{current_iteration}_results>"
                    if tool_result.document_results:
                        results_data += f"\n<document_references>\n{yaml.dump(tool_result.document_results, allow_unicode=True, sort_keys=False, default_flow_style=False)}\n</document_references>"
                    if tool_result.online_results:
                        results_data += f"\n<online_results>\n{yaml.dump(tool_result.online_results, allow_unicode=True, sort_keys=False, default_flow_style=False)}\n</online_results>"
                    if tool_result.code_results:
                        results_data += f"\n<code_results>\n{yaml.dump(truncate_code_context(tool_result.code_results), allow_unicode=True, sort_keys=False, default_flow_style=False)}\n</code_results>"
                    if tool_result.operator_results:
                        results_data += f"\n<browser_operator_results>\n{tool_result.operator_results.response}\n</browser_operator_results>"
                    if tool_result.mcp_results:
                        results_data += f"\n<mcp_tool_results>\n{yaml.dump(tool_result.mcp_results, allow_unicode=True, sort_keys=False, default_flow_style=False)}\n</mcp_tool_results>"
                    if this_iteration.warning:
                        results_data += f"\n<warning>\n{this_iteration.warning}\n</warning>"
                    results_data += f"\n</iteration_{current_iteration}_results>"

                    this_iteration.summarizedResult = results_data

                this_iteration.summarizedResult = (
                    this_iteration.summarizedResult
                    or f"<iteration_{current_iteration}_results>Failed to get results.</iteration_{current_iteration}_results>"
                )
                previous_iterations.append(this_iteration)
                yield this_iteration

    # Close MCP client connections
    for mcp_client in mcp_clients:
        await mcp_client.close()



# System Packages
import json

from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import FileResponse, HTMLResponse, RedirectResponse
from fastapi.templating import Jinja2Templates
from starlette.authentication import requires

from khoj.database.adapters import get_user_github_config
from khoj.routers.helpers import get_next_url, get_user_config
from khoj.utils import constants, state
from khoj.utils.rawconfig import GithubContentConfig, GithubRepoConfig

# Initialize Router
web_client = APIRouter()
templates = Jinja2Templates([constants.web_directory, constants.next_js_directory, constants.pypi_static_directory])
home_templates = Jinja2Templates([constants.home_directory])


# Create Routes
@web_client.get("/", response_class=FileResponse)
def index(request: Request):
    # Redirect unauthenticated users to /home landing page when not in anonymous mode
    # Skip redirect if user explicitly navigated from home page (indicated by query param)
    if not state.anonymous_mode and not request.user.is_authenticated:
        if "v" not in request.query_params:
            return RedirectResponse(url="/home")
    return templates.TemplateResponse("index.html", context={"request": request})


@web_client.post("/", response_class=FileResponse)
@requires(["authenticated"], redirect="login_page")
def index_post(request: Request):
    return templates.TemplateResponse("index.html", context={"request": request})


@web_client.get("/home", response_class=HTMLResponse)
def home_page(request: Request):
    """Serve the landing page for unauthenticated users"""
    # If user is authenticated, redirect to main app
    if request.user.is_authenticated:
        return RedirectResponse(url="/")
    return home_templates.TemplateResponse("index.html", context={"request": request})


@web_client.get("/home/{file_path:path}", response_class=FileResponse)
def home_static_files(file_path: str):
    """Serve static files from the home landing page directory"""
    resolved = (constants.home_directory / file_path).resolve()
    if not resolved.is_relative_to(constants.home_directory.resolve()):
        raise HTTPException(status_code=404)
    return FileResponse(resolved)


@web_client.get("/search", response_class=FileResponse)
@requires(["authenticated"], redirect="login_page")
def search_page(request: Request):
    return templates.TemplateResponse(
        "search/index.html",
        context={
            "request": request,
        },
    )


@web_client.get("/chat", response_class=FileResponse)
@requires(["authenticated"], redirect="login_page")
def chat_page(request: Request):
    return templates.TemplateResponse(
        "chat/index.html",
        context={
            "request": request,
        },
    )


@web_client.get("/login", response_class=FileResponse)
def login_page(request: Request):
    next_url = get_next_url(request)
    if request.user.is_authenticated:
        return RedirectResponse(url=next_url)
    # Redirect to main app which shows the login popup for unauthenticated users
    # Append v=app to prevent redirect loop back to /home
    redirect_url = f"/?v=app&next={next_url}" if next_url != "/" else "/?v=app"
    return RedirectResponse(url=redirect_url)


@web_client.get("/agents", response_class=HTMLResponse)
def agents_page(request: Request):
    return templates.TemplateResponse(
        "agents/index.html",
        context={
            "request": request,
        },
    )


@web_client.get("/settings", response_class=HTMLResponse)
@requires(["authenticated"], redirect="login_page")
def config_page(request: Request):
    return templates.TemplateResponse("settings/index.html", context={"request": request})


@web_client.get("/settings/content/github", response_class=HTMLResponse)
@requires(["authenticated"], redirect="login_page")
def github_config_page(request: Request):
    user = request.user.object
    user_config = get_user_config(user, request)
    current_github_config = get_user_github_config(user)

    if current_github_config:
        raw_repos = current_github_config.githubrepoconfig.all()
        repos = []
        for repo in raw_repos:
            repos.append(
                GithubRepoConfig(
                    name=repo.name,
                    owner=repo.owner,
                    branch=repo.branch,
                )
            )
        current_config = GithubContentConfig(
            pat_token=current_github_config.pat_token,
            repos=repos,
        )
        current_config = json.loads(current_config.json())
    else:
        current_config = {}  # type: ignore

    user_config["current_config"] = current_config
    return templates.TemplateResponse("content_source_github_input.html", context=user_config)


@web_client.get("/share/chat/{public_conversation_slug}", response_class=HTMLResponse)
def view_public_conversation(request: Request):
    return templates.TemplateResponse(
        "share/chat/index.html",
        context={
            "request": request,
        },
    )


@web_client.get("/automations", response_class=HTMLResponse)
def automations_config_page(
    request: Request,
):
    return templates.TemplateResponse(
        "automations/index.html",
        context={
            "request": request,
        },
    )


@web_client.get("/.well-known/assetlinks.json", response_class=FileResponse)
def assetlinks(request: Request):
    return FileResponse(constants.assetlinks_file_path)


@web_client.get("/server/error", response_class=HTMLResponse)
def server_error_page(request: Request):
    return templates.TemplateResponse("error.html", context={"request": request})



import logging
from contextlib import AsyncExitStack
from typing import Optional

from mcp import ClientSession, StdioServerParameters
from mcp.client.sse import sse_client
from mcp.client.stdio import stdio_client
from mcp.types import AudioContent, ImageContent, TextContent

logger = logging.getLogger(__name__)


class MCPClient:
    """
    A client for interacting with MCP servers.
    Establishes a session with the server and provides methods to get and run tools.
    Supports both stdio and sse communication methods.
    """

    def __init__(self, name: str, path: str, api_key: Optional[str] = None):
        self.name = name
        self.path = path
        self.api_key = api_key
        self.session: Optional[ClientSession] = None
        self.exit_stack = AsyncExitStack()

    async def connect(self):
        """
        Connect to the MCP server based on the specified path.
        Uses stdio for local scripts and SSE for remote servers.
        """
        if self.path.startswith("http://") or self.path.startswith("https://"):
            # Path is a URL, so use SSE to connect to the server
            await self._connect_to_sse_server()
        else:
            # Path is a script, so use stdio to connect to the server
            await self._connect_to_stdio_server()

    async def get_tools(self):
        """
        Retrieve the list of tools available on the MCP server.
        """
        # Ensure connection is established
        if not self.session:
            await self.connect()

        tools_response = await self.session.list_tools()
        return tools_response.tools

    async def run_tool(self, tool_name: str, input_data: dict):
        """
        Run a specified tool on the MCP server with the given input data.
        """
        # Ensure connection is established
        if not self.session:
            await self.connect()

        result = await self.session.call_tool(tool_name, input_data)

        # Process result content based on its type
        if len(result.content) > 0 and isinstance(result.content[0], TextContent):
            return [item.text for item in result.content]
        elif len(result.content) > 0 and isinstance(result.content[0], AudioContent):
            return [{"data": item.data, "format": item.mimeType} for item in result.content]
        elif len(result.content) > 0 and isinstance(result.content[0], ImageContent):
            return [{"data": item.data, "format": item.mimeType} for item in result.content]
        return result.content

    async def _connect_to_sse_server(self):
        """
        Connect to the MCP server using Server-Sent Events (SSE).
        """
        self._streams_context = sse_client(url=self.path)
        streams = await self._streams_context.__aenter__()

        self._session_context = ClientSession(*streams)
        self.session = await self._session_context.__aenter__()

        # Initialize
        await self.session.initialize()

    async def _connect_to_stdio_server(self):
        """
        Connect to the MCP server using stdio communication.
        """
        is_python = False
        is_javascript = False
        command = None
        args = [self.path]

        # Determine if the server is a file path or npm package
        if self.path.startswith("@") or "/" not in self.path:
            # Assume it's an npm package
            is_javascript = True
            command = "npx"
        else:
            # It's a file path
            is_python = self.path.endswith(".py")
            is_javascript = self.path.endswith(".js")
            if not (is_python or is_javascript):
                raise ValueError("Server script must be a .py, .js file or npm package.")

            command = "python" if is_python else "node"

        server_params = StdioServerParameters(command=command, args=args, env=None)

        logger.debug(f"Connecting to stdio MCP server with command: {command} and args: {args}")

        # Start the server
        stdio_transport = await self.exit_stack.enter_async_context(stdio_client(server_params))
        self.stdio, self.writer = stdio_transport
        self.session = await self.exit_stack.enter_async_context(ClientSession(self.stdio, self.writer))

        await self.session.initialize()

    async def close(self):
        """
        Close the MCP client session and clean up resources.
        """
        await self.exit_stack.aclose()
        if hasattr(self, "_session_context") and self._session_context:
            await self._session_context.__aexit__(None, None, None)
        if hasattr(self, "_streams_context") and self._streams_context:
            await self._streams_context.__aexit__(None, None, None)



# Generated by Django 5.0.10 on 2025-01-21 18:42
from typing import List

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0081_agent_is_hidden_chatmodel_description_and_more"),
        ("database", "0081_merge_20250120_1633"),
    ]

    operations: List[str] = []



import logging
import os
import re
import uuid
from random import choice
from typing import Dict, List, Optional, Union

from django.contrib.auth.models import AbstractUser
from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import pre_save
from django.dispatch import receiver
from pgvector.django import VectorField
from phonenumber_field.modelfields import PhoneNumberField
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, model_validator

logger = logging.getLogger(__name__)


# Pydantic models for type Chat Message validation
class Context(PydanticBaseModel):
    compiled: str
    file: str
    uri: Optional[str] = None
    query: Optional[str] = None

    @model_validator(mode="after")
    def set_uri_fallback(self):
        """Set the URI to existing deeplink URI. Fallback to file based URI if unset."""
        if self.uri and self.uri.strip():
            self.uri = self.uri
        elif self.file and (self.file.startswith("http") or self.file.startswith("file://")):
            self.uri = self.file
        elif self.file:
            self.uri = f"file://{self.file}"
        else:
            self.uri = None
        return self


class CodeContextFile(PydanticBaseModel):
    filename: str
    b64_data: str


class CodeContextResult(PydanticBaseModel):
    success: bool
    output_files: List[CodeContextFile]
    std_out: Optional[str] = None
    std_err: str
    code_runtime: Optional[int] = None


class CodeContextData(PydanticBaseModel):
    code: str
    results: Optional[CodeContextResult] = None


class WebPage(PydanticBaseModel):
    link: str
    query: Optional[str] = None
    snippet: str


class AnswerBox(PydanticBaseModel):
    link: Optional[str] = None
    snippet: Optional[str] = None
    title: str
    snippetHighlighted: Optional[List[str]] = None


class PeopleAlsoAsk(PydanticBaseModel):
    link: Optional[str] = None
    question: Optional[str] = None
    snippet: Optional[str] = None
    title: Optional[str] = None


class KnowledgeGraph(PydanticBaseModel):
    attributes: Optional[Dict[str, str]] = None
    description: Optional[str] = None
    descriptionLink: Optional[str] = None
    descriptionSource: Optional[str] = None
    imageUrl: Optional[str] = None
    title: str
    type: Optional[str] = None


class OrganicContext(PydanticBaseModel):
    snippet: Optional[str] = None
    title: str
    link: str


class OnlineContext(PydanticBaseModel):
    webpages: Optional[Union[WebPage, List[WebPage]]] = None
    answerBox: Optional[AnswerBox] = None
    peopleAlsoAsk: Optional[List[PeopleAlsoAsk]] = None
    knowledgeGraph: Optional[KnowledgeGraph] = None
    organic: Optional[List[OrganicContext]] = None


class Intent(PydanticBaseModel):
    type: str
    query: Optional[str] = None
    memory_type: Optional[str] = Field(alias="memory-type", default=None)
    inferred_queries: Optional[List[str]] = Field(default=None, alias="inferred-queries")


class TrainOfThought(PydanticBaseModel):
    type: str
    data: str


class ChatMessageModel(PydanticBaseModel):
    by: str
    message: str | list[dict]
    trainOfThought: List[TrainOfThought] = []
    context: List[Context] = []
    onlineContext: Dict[str, OnlineContext] = {}
    codeContext: Dict[str, CodeContextData] = {}
    researchContext: Optional[List] = None
    operatorContext: Optional[List] = None
    created: Optional[str] = None
    images: Optional[List[str]] = None
    queryFiles: Optional[List[Dict]] = None
    excalidrawDiagram: Optional[List[Dict]] = None
    mermaidjsDiagram: Optional[str] = None
    turnId: Optional[str] = None
    intent: Optional[Intent] = None
    automationId: Optional[str] = None


class DbBaseModel(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class ClientApplication(DbBaseModel):
    name = models.CharField(max_length=200)
    client_id = models.CharField(max_length=200)
    client_secret = models.CharField(max_length=200)

    def __str__(self):
        return self.name


class KhojUser(AbstractUser):
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)
    phone_number = PhoneNumberField(null=True, default=None, blank=True)
    verified_phone_number = models.BooleanField(default=False)
    verified_email = models.BooleanField(default=False)
    email_verification_code = models.CharField(max_length=200, null=True, default=None, blank=True)
    email_verification_code_expiry = models.DateTimeField(null=True, default=None, blank=True)

    def save(self, *args, **kwargs):
        if not self.uuid:
            self.uuid = uuid.uuid4()
        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.username} ({self.uuid})"


class GoogleUser(models.Model):
    user = models.OneToOneField(KhojUser, on_delete=models.CASCADE)
    sub = models.CharField(max_length=200)
    azp = models.CharField(max_length=200)
    email = models.CharField(max_length=200)
    name = models.CharField(max_length=200, null=True, default=None, blank=True)
    given_name = models.CharField(max_length=200, null=True, default=None, blank=True)
    family_name = models.CharField(max_length=200, null=True, default=None, blank=True)
    picture = models.CharField(max_length=200, null=True, default=None)
    locale = models.CharField(max_length=200, null=True, default=None, blank=True)

    def __str__(self):
        return self.name


class KhojApiUser(models.Model):
    """User issued API tokens to authenticate Khoj clients"""

    user = models.ForeignKey(KhojUser, on_delete=models.CASCADE)
    token = models.CharField(max_length=50, unique=True)
    name = models.CharField(max_length=50)
    accessed_at = models.DateTimeField(null=True, default=None)


class Subscription(DbBaseModel):
    class Type(models.TextChoices):
        TRIAL = "trial"
        STANDARD = "standard"

    user = models.OneToOneField(KhojUser, on_delete=models.CASCADE, related_name="subscription")
    type = models.CharField(max_length=20, choices=Type.choices, default=Type.STANDARD)
    is_recurring = models.BooleanField(default=False)
    renewal_date = models.DateTimeField(null=True, default=None, blank=True)
    enabled_trial_at = models.DateTimeField(null=True, default=None, blank=True)


class AiModelApi(DbBaseModel):
    name = models.CharField(max_length=200)
    api_key = models.CharField(max_length=4000)
    api_base_url = models.URLField(max_length=200, default=None, blank=True, null=True)

    def __str__(self):
        return self.name


class PriceTier(models.TextChoices):
    FREE = "free"
    STANDARD = "standard"


class ChatModel(DbBaseModel):
    class ModelType(models.TextChoices):
        OPENAI = "openai"
        ANTHROPIC = "anthropic"
        GOOGLE = "google"

    max_prompt_size = models.IntegerField(default=None, null=True, blank=True)
    subscribed_max_prompt_size = models.IntegerField(default=None, null=True, blank=True)
    tokenizer = models.CharField(max_length=200, default=None, null=True, blank=True)
    name = models.CharField(max_length=200, default="gemini-2.5-flash")
    friendly_name = models.CharField(max_length=200, default=None, null=True, blank=True)
    model_type = models.CharField(max_length=200, choices=ModelType.choices, default=ModelType.GOOGLE)
    price_tier = models.CharField(max_length=20, choices=PriceTier.choices, default=PriceTier.FREE)
    vision_enabled = models.BooleanField(default=False)
    ai_model_api = models.ForeignKey(AiModelApi, on_delete=models.CASCADE, default=None, null=True, blank=True)
    description = models.TextField(default=None, null=True, blank=True)
    strengths = models.TextField(default=None, null=True, blank=True)

    def __str__(self):
        return self.friendly_name or self.name


class VoiceModelOption(DbBaseModel):
    model_id = models.CharField(max_length=200)
    name = models.CharField(max_length=200)
    price_tier = models.CharField(max_length=20, choices=PriceTier.choices, default=PriceTier.STANDARD)


class Agent(DbBaseModel):
    class StyleColorTypes(models.TextChoices):
        BLUE = "blue"
        GREEN = "green"
        RED = "red"
        YELLOW = "yellow"
        ORANGE = "orange"
        PURPLE = "purple"
        PINK = "pink"
        TEAL = "teal"
        CYAN = "cyan"
        LIME = "lime"
        INDIGO = "indigo"
        FUCHSIA = "fuchsia"
        ROSE = "rose"
        SKY = "sky"
        AMBER = "amber"
        EMERALD = "emerald"

    class StyleIconTypes(models.TextChoices):
        LIGHTBULB = "Lightbulb"
        HEALTH = "Health"
        ROBOT = "Robot"
        APERTURE = "Aperture"
        GRADUATION_CAP = "GraduationCap"
        JEEP = "Jeep"
        ISLAND = "Island"
        MATH_OPERATIONS = "MathOperations"
        ASCLEPIUS = "Asclepius"
        COUCH = "Couch"
        CODE = "Code"
        ATOM = "Atom"
        CLOCK_COUNTER_CLOCKWISE = "ClockCounterClockwise"
        PENCIL_LINE = "PencilLine"
        CHALKBOARD = "Chalkboard"
        CIGARETTE = "Cigarette"
        CRANE_TOWER = "CraneTower"
        HEART = "Heart"
        LEAF = "Leaf"
        NEWSPAPER_CLIPPING = "NewspaperClipping"
        ORANGE_SLICE = "OrangeSlice"
        SMILEY_MELTING = "SmileyMelting"
        YIN_YANG = "YinYang"
        SNEAKER_MOVE = "SneakerMove"
        STUDENT = "Student"
        OVEN = "Oven"
        GAVEL = "Gavel"
        BROADCAST = "Broadcast"

    class PrivacyLevel(models.TextChoices):
        PUBLIC = "public"
        PRIVATE = "private"
        PROTECTED = "protected"

    class InputToolOptions(models.TextChoices):
        # These map to various ConversationCommand types
        GENERAL = "general"
        ONLINE = "online"
        NOTES = "notes"
        WEBPAGE = ("webpage",)
        CODE = "code"

    class OutputModeOptions(models.TextChoices):
        # These map to various ConversationCommand types
        IMAGE = "image"
        DIAGRAM = "diagram"

    creator = models.ForeignKey(
        KhojUser, on_delete=models.CASCADE, default=None, null=True, blank=True
    )  # Creator will only be null when the agents are managed by admin
    name = models.CharField(max_length=200)
    personality = models.TextField(default=None, null=True, blank=True)
    input_tools = ArrayField(
        models.CharField(max_length=200, choices=InputToolOptions.choices), default=list, null=True, blank=True
    )
    output_modes = ArrayField(
        models.CharField(max_length=200, choices=OutputModeOptions.choices), default=list, null=True, blank=True
    )
    managed_by_admin = models.BooleanField(default=False)
    chat_model = models.ForeignKey(ChatModel, on_delete=models.CASCADE)
    slug = models.CharField(max_length=200, unique=True)
    style_color = models.CharField(max_length=200, choices=StyleColorTypes.choices, default=StyleColorTypes.ORANGE)
    style_icon = models.CharField(max_length=200, choices=StyleIconTypes.choices, default=StyleIconTypes.LIGHTBULB)
    privacy_level = models.CharField(max_length=30, choices=PrivacyLevel.choices, default=PrivacyLevel.PRIVATE)
    is_hidden = models.BooleanField(default=False)

    def save(self, *args, **kwargs):
        is_new = self._state.adding

        if self.creator is None:
            self.managed_by_admin = True

        if is_new and not self.slug:
            random_sequence = "".join(choice("0123456789") for i in range(6))
            slug = f"{self.name.lower().replace(' ', '-')}-{random_sequence}"
            self.slug = slug

        super().save(*args, **kwargs)

    def __str__(self):
        return self.name


class ProcessLock(DbBaseModel):
    class Operation(models.TextChoices):
        INDEX_CONTENT = "index_content"
        SCHEDULED_JOB = "scheduled_job"
        SCHEDULE_LEADER = "schedule_leader"
        APPLY_MIGRATIONS = "apply_migrations"

    # We need to make sure that some operations are thread-safe. To do so, add locks for potentially shared operations.
    # For example, we need to make sure that only one process is updating the embeddings at a time.
    name = models.CharField(max_length=200, choices=Operation.choices, unique=True)
    started_at = models.DateTimeField(auto_now_add=True)
    max_duration_in_seconds = models.IntegerField(default=60 * 60 * 12)  # 12 hours


@receiver(pre_save, sender=Agent)
def verify_agent(sender, instance, **kwargs):
    # check if this is a new instance
    if instance._state.adding:
        if Agent.objects.filter(name=instance.name, privacy_level=Agent.PrivacyLevel.PUBLIC).exists():
            raise ValidationError(f"A public Agent with the name {instance.name} already exists.")
        if Agent.objects.filter(name=instance.name, creator=instance.creator).exists():
            raise ValidationError(f"A private Agent with the name {instance.name} already exists.")


class NotionConfig(DbBaseModel):
    token = models.CharField(max_length=200)
    user = models.ForeignKey(KhojUser, on_delete=models.CASCADE)


class GithubConfig(DbBaseModel):
    pat_token = models.CharField(max_length=200)
    user = models.ForeignKey(KhojUser, on_delete=models.CASCADE)


class GithubRepoConfig(DbBaseModel):
    name = models.CharField(max_length=200)
    owner = models.CharField(max_length=200)
    branch = models.CharField(max_length=200)
    github_config = models.ForeignKey(GithubConfig, on_delete=models.CASCADE, related_name="githubrepoconfig")


class WebScraper(DbBaseModel):
    class WebScraperType(models.TextChoices):
        FIRECRAWL = "Firecrawl"
        OLOSTEP = "Olostep"
        EXA = "Exa"
        DIRECT = "Direct"

    name = models.CharField(
        max_length=200,
        default=None,
        null=True,
        blank=True,
        unique=True,
        help_text="Friendly name. If not set, it will be set to the type of the scraper.",
    )
    type = models.CharField(max_length=20, choices=WebScraperType.choices, default=WebScraperType.DIRECT)
    api_key = models.CharField(
        max_length=200,
        default=None,
        null=True,
        blank=True,
        help_text="API key of the web scraper. Only set if scraper service requires an API key. Default is set from env var.",
    )
    api_url = models.URLField(
        max_length=200,
        default=None,
        null=True,
        blank=True,
        help_text="API URL of the web scraper. Only set if scraper service on non-default URL.",
    )
    priority = models.IntegerField(
        default=None,
        null=True,
        blank=True,
        unique=True,
        help_text="Priority of the web scraper. Lower numbers run first.",
    )

    def clean(self):
        error = {}
        if self.name is None:
            self.name = self.type.capitalize()
        if self.api_url is None:
            if self.type == self.WebScraperType.FIRECRAWL:
                self.api_url = os.getenv("FIRECRAWL_API_URL", "https://api.firecrawl.dev")
            elif self.type == self.WebScraperType.OLOSTEP:
                self.api_url = os.getenv("OLOSTEP_API_URL", "https://agent.olostep.com/olostep-p2p-incomingAPI")
            elif self.type == self.WebScraperType.EXA:
                self.api_url = os.getenv("EXA_API_URL", "https://api.exa.ai")
        if self.api_key is None:
            if self.type == self.WebScraperType.FIRECRAWL:
                self.api_key = os.getenv("FIRECRAWL_API_KEY")
                if not self.api_key and self.api_url == "https://api.firecrawl.dev":
                    error["api_key"] = "Set API key to use default Firecrawl. Get API key from https://firecrawl.dev."
            elif self.type == self.WebScraperType.OLOSTEP:
                self.api_key = os.getenv("OLOSTEP_API_KEY")
                if self.api_key is None:
                    error["api_key"] = "Set API key to use Olostep. Get API key from https://olostep.com/."
            elif self.type == self.WebScraperType.EXA:
                self.api_key = os.getenv("EXA_API_KEY")
                if self.api_key is None:
                    error["api_key"] = "Set API key to use Exa. Get API key from https://exa.ai/."
        if error:
            raise ValidationError(error)

    def save(self, *args, **kwargs):
        self.clean()

        if self.priority is None:
            max_priority = WebScraper.objects.aggregate(models.Max("priority"))["priority__max"]
            self.priority = max_priority + 1 if max_priority else 1

        super().save(*args, **kwargs)

    def __str__(self):
        return self.name


class ServerChatSettings(DbBaseModel):
    class ChatModelSlot(models.TextChoices):
        """Enum for the different chat model slots in ServerChatSettings"""

        CHAT_DEFAULT = "chat_default"
        CHAT_ADVANCED = "chat_advanced"
        THINK_FREE_FAST = "think_free_fast"
        THINK_FREE_DEEP = "think_free_deep"
        THINK_PAID_FAST = "think_paid_fast"
        THINK_PAID_DEEP = "think_paid_deep"

    class MemoryMode(models.TextChoices):
        """Enum for server-level memory feature configuration"""

        DISABLED = "disabled", "Disabled"
        ENABLED_DEFAULT_OFF = "enabled_default_off", "Enabled, default off"
        ENABLED_DEFAULT_ON = "enabled_default_on", "Enabled, default on"

    chat_default = models.ForeignKey(
        ChatModel, on_delete=models.CASCADE, default=None, null=True, blank=True, related_name="chat_default"
    )
    chat_advanced = models.ForeignKey(
        ChatModel, on_delete=models.CASCADE, default=None, null=True, blank=True, related_name="chat_advanced"
    )
    think_free_fast = models.ForeignKey(
        ChatModel, on_delete=models.CASCADE, default=None, null=True, blank=True, related_name="think_free_fast"
    )
    think_free_deep = models.ForeignKey(
        ChatModel, on_delete=models.CASCADE, default=None, null=True, blank=True, related_name="think_free_deep"
    )
    think_paid_fast = models.ForeignKey(
        ChatModel, on_delete=models.CASCADE, default=None, null=True, blank=True, related_name="think_paid_fast"
    )
    think_paid_deep = models.ForeignKey(
        ChatModel, on_delete=models.CASCADE, default=None, null=True, blank=True, related_name="think_paid_deep"
    )
    web_scraper = models.ForeignKey(
        WebScraper, on_delete=models.CASCADE, default=None, null=True, blank=True, related_name="web_scraper"
    )
    priority = models.IntegerField(
        default=None,
        null=True,
        blank=True,
        unique=True,
        help_text="Priority of the server chat settings. Lower numbers run first.",
    )
    memory_mode = models.CharField(
        max_length=20,
        choices=MemoryMode.choices,
        default=MemoryMode.ENABLED_DEFAULT_ON,
        help_text="Server-level memory feature configuration. Disabled overrides user preference.",
    )

    def clean(self):
        error = {}
        if self.chat_default and self.chat_default.price_tier != PriceTier.FREE:
            error["chat_default"] = "Set the price tier of this chat model to free or use a free tier chat model."
        if self.think_free_fast and self.think_free_fast.price_tier != PriceTier.FREE:
            error["think_free_fast"] = "Set the price tier of this chat model to free or use a free tier chat model."
        if self.think_free_deep and self.think_free_deep.price_tier != PriceTier.FREE:
            error["think_free_deep"] = "Set the price tier of this chat model to free or use a free tier chat model."
        if error:
            raise ValidationError(error)

    def save(self, *args, **kwargs):
        self.clean()

        if self.priority is None:
            max_priority = ServerChatSettings.objects.aggregate(models.Max("priority"))["priority__max"]
            self.priority = max_priority + 1 if max_priority else 1

        super().save(*args, **kwargs)


class SearchModelConfig(DbBaseModel):
    class ModelType(models.TextChoices):
        TEXT = "text"

    class ApiType(models.TextChoices):
        HUGGINGFACE = "huggingface"
        OPENAI = "openai"
        LOCAL = "local"

    # This is the model name exposed to users on their settings page
    name = models.CharField(max_length=200, default="default")
    # Type of content the model can generate embeddings for
    model_type = models.CharField(max_length=200, choices=ModelType.choices, default=ModelType.TEXT)
    # Bi-encoder model of sentence-transformer type to load from HuggingFace
    bi_encoder = models.CharField(max_length=200, default="thenlper/gte-small")
    # Config passed to the sentence-transformer model constructor. E.g. device="cuda:0", trust_remote_server=True etc.
    bi_encoder_model_config = models.JSONField(default=dict, blank=True)
    # Query encode configs like prompt, precision, normalize_embeddings, etc. for sentence-transformer models
    bi_encoder_query_encode_config = models.JSONField(default=dict, blank=True)
    # Docs encode configs like prompt, precision, normalize_embeddings, etc. for sentence-transformer models
    bi_encoder_docs_encode_config = models.JSONField(default=dict, blank=True)
    # Cross-encoder model of sentence-transformer type to load from HuggingFace
    cross_encoder = models.CharField(max_length=200, default="mixedbread-ai/mxbai-rerank-xsmall-v1")
    # Config passed to the cross-encoder model constructor. E.g. device="cuda:0", trust_remote_server=True etc.
    cross_encoder_model_config = models.JSONField(default=dict, blank=True)
    # Inference server API endpoint to use for embeddings inference. Bi-encoder model should be hosted on this server
    embeddings_inference_endpoint = models.CharField(max_length=200, default=None, null=True, blank=True)
    # Inference server API Key to use for embeddings inference. Bi-encoder model should be hosted on this server
    embeddings_inference_endpoint_api_key = models.CharField(max_length=200, default=None, null=True, blank=True)
    # Inference server API type to use for embeddings inference.
    embeddings_inference_endpoint_type = models.CharField(
        max_length=200, choices=ApiType.choices, default=ApiType.LOCAL
    )
    # Inference server API endpoint to use for embeddings inference. Cross-encoder model should be hosted on this server
    cross_encoder_inference_endpoint = models.CharField(max_length=200, default=None, null=True, blank=True)
    # Inference server API Key to use for embeddings inference. Cross-encoder model should be hosted on this server
    cross_encoder_inference_endpoint_api_key = models.CharField(max_length=200, default=None, null=True, blank=True)
    # The confidence threshold of the bi_encoder model to consider the embeddings as relevant
    bi_encoder_confidence_threshold = models.FloatField(default=0.18)

    def __str__(self):
        return self.name


class TextToImageModelConfig(DbBaseModel):
    class ModelType(models.TextChoices):
        OPENAI = "openai"
        REPLICATE = "replicate"
        GOOGLE = "google"

    model_name = models.CharField(max_length=200, default="dall-e-3")
    friendly_name = models.CharField(max_length=200, default=None, null=True, blank=True)
    model_type = models.CharField(max_length=200, choices=ModelType.choices, default=ModelType.OPENAI)
    price_tier = models.CharField(max_length=20, choices=PriceTier.choices, default=PriceTier.FREE)
    api_key = models.CharField(max_length=200, default=None, null=True, blank=True)
    ai_model_api = models.ForeignKey(AiModelApi, on_delete=models.CASCADE, default=None, null=True, blank=True)

    def clean(self):
        # Custom validation logic
        error = {}
        if self.model_type == self.ModelType.OPENAI:
            if self.api_key and self.ai_model_api:
                error["api_key"] = (
                    "Both API key and AI Model API cannot be set for OpenAI models. Please set only one of them."
                )
                error["ai_model_api"] = (
                    "Both API key and OpenAI config cannot be set for OpenAI models. Please set only one of them."
                )
        if self.model_type != self.ModelType.OPENAI and self.model_type != self.ModelType.GOOGLE:
            if not self.api_key:
                error["api_key"] = "The API key field must be set for non OpenAI, non Google models."
            if self.ai_model_api:
                error["ai_model_api"] = "AI Model API cannot be set for non OpenAI, non Google models."
        if error:
            raise ValidationError(error)

    def save(self, *args, **kwargs):
        self.clean()
        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.model_name} - {self.model_type}"


class SpeechToTextModelOptions(DbBaseModel):
    class ModelType(models.TextChoices):
        OPENAI = "openai"

    model_name = models.CharField(max_length=200, default="whisper-1")
    friendly_name = models.CharField(max_length=200, default=None, null=True, blank=True)
    model_type = models.CharField(max_length=200, choices=ModelType.choices, default=ModelType.OPENAI)
    price_tier = models.CharField(max_length=20, choices=PriceTier.choices, default=PriceTier.FREE)
    ai_model_api = models.ForeignKey(AiModelApi, on_delete=models.CASCADE, default=None, null=True, blank=True)

    def __str__(self):
        return f"{self.model_name} - {self.model_type}"


class UserConversationConfig(DbBaseModel):
    user = models.OneToOneField(KhojUser, on_delete=models.CASCADE)
    setting = models.ForeignKey(ChatModel, on_delete=models.CASCADE, default=None, null=True, blank=True)
    enable_memory = models.BooleanField(default=True)


class UserVoiceModelConfig(DbBaseModel):
    user = models.OneToOneField(KhojUser, on_delete=models.CASCADE)
    setting = models.ForeignKey(VoiceModelOption, on_delete=models.CASCADE, default=None, null=True, blank=True)


class UserTextToImageModelConfig(DbBaseModel):
    user = models.OneToOneField(KhojUser, on_delete=models.CASCADE)
    setting = models.ForeignKey(TextToImageModelConfig, on_delete=models.CASCADE)


class Conversation(DbBaseModel):
    user = models.ForeignKey(KhojUser, on_delete=models.CASCADE)
    conversation_log = models.JSONField(default=dict)
    client = models.ForeignKey(ClientApplication, on_delete=models.CASCADE, default=None, null=True, blank=True)

    # Slug is an app-generated conversation identifier. Need not be unique. Used as display title essentially.
    slug = models.CharField(max_length=200, default=None, null=True, blank=True)

    # The title field is explicitly set by the user.
    title = models.CharField(max_length=500, default=None, null=True, blank=True)
    agent = models.ForeignKey(Agent, on_delete=models.SET_NULL, default=None, null=True, blank=True)
    file_filters = models.JSONField(default=list)
    id = models.UUIDField(default=uuid.uuid4, editable=False, unique=True, primary_key=True, db_index=True)

    def clean(self):
        # Validate conversation_log structure
        try:
            messages = self.conversation_log.get("chat", [])
            for msg in messages:
                ChatMessageModel.model_validate(msg)
        except Exception as e:
            raise ValidationError(f"Invalid conversation_log format: {str(e)}")

    def save(self, *args, **kwargs):
        self.clean()
        super().save(*args, **kwargs)

    @property
    def messages(self) -> List[ChatMessageModel]:
        """Type-hinted accessor for conversation messages"""
        validated_messages = []
        for msg in self.conversation_log.get("chat", []):
            try:
                # Clean up inferred queries if they contain None
                if msg.get("intent") and msg["intent"].get("inferred_queries"):
                    msg["intent"]["inferred-queries"] = [
                        q for q in msg["intent"]["inferred_queries"] if q is not None and isinstance(q, str)
                    ]
                msg["message"] = str(msg.get("message", ""))
                validated_messages.append(ChatMessageModel.model_validate(msg))
            except ValidationError as e:
                logger.warning(f"Skipping invalid message in conversation: {e}")
                continue
        return validated_messages

    async def pop_message(self, interrupted: bool = False) -> Optional[ChatMessageModel]:
        """
        Remove and return the last message from the conversation log, persisting the change to the database.
        When interrupted is True, we only drop the last message if it was an interrupted message by khoj.
        """
        chat_log = self.conversation_log.get("chat", [])

        if not chat_log:
            return None

        last_message = chat_log[-1]
        is_interrupted_msg = last_message.get("by") == "khoj" and not last_message.get("message")
        # When handling an interruption, only pop if the last message is an empty one by khoj.
        if interrupted and not is_interrupted_msg:
            return None

        # Pop the last message, save the conversation, and then return the message.
        popped_message_dict = chat_log.pop()
        await self.asave()

        # Try to validate and return the popped message as a Pydantic model
        try:
            return ChatMessageModel.model_validate(popped_message_dict)
        except ValidationError as e:
            logger.warning(f"Popped an invalid message from conversation. The removal has been saved. Error: {e}")
            # The invalid message was removed and saved, but we can't return a valid model.
            return None


class PublicConversation(DbBaseModel):
    source_owner = models.ForeignKey(KhojUser, on_delete=models.CASCADE)
    conversation_log = models.JSONField(default=dict)
    slug = models.CharField(max_length=200, default=None, null=True, blank=True)
    title = models.CharField(max_length=200, default=None, null=True, blank=True)
    agent = models.ForeignKey(Agent, on_delete=models.SET_NULL, default=None, null=True, blank=True)


@receiver(pre_save, sender=PublicConversation)
def verify_public_conversation(sender, instance, **kwargs):
    # check if this is a new instance
    if instance._state.adding:
        base_length = 50  # Base slug length before adding random suffix
        base_slug = re.sub(r"\W+", "-", instance.slug.lower())[:base_length] if instance.slug else uuid.uuid4().hex
        suffix_length = 8  # Length of the random suffix to ensure uniqueness
        while True:
            random_id = uuid.uuid4().hex[:suffix_length]
            slug = f"{base_slug}-{random_id}"
            if not PublicConversation.objects.filter(slug=slug).exists():
                break
        instance.slug = slug


class ReflectiveQuestion(DbBaseModel):
    question = models.CharField(max_length=500)
    user = models.ForeignKey(KhojUser, on_delete=models.CASCADE, default=None, null=True, blank=True)


class FileObject(DbBaseModel):
    # Contains the full text of a file that has associated Entry objects
    file_name = models.CharField(max_length=400, default=None, null=True, blank=True)
    raw_text = models.TextField()
    user = models.ForeignKey(KhojUser, on_delete=models.CASCADE, default=None, null=True, blank=True)
    agent = models.ForeignKey(Agent, on_delete=models.CASCADE, default=None, null=True, blank=True)


class Entry(DbBaseModel):
    class EntryType(models.TextChoices):
        IMAGE = "image"
        PDF = "pdf"
        PLAINTEXT = "plaintext"
        MARKDOWN = "markdown"
        ORG = "org"
        NOTION = "notion"
        GITHUB = "github"
        CONVERSATION = "conversation"
        DOCX = "docx"

    class EntrySource(models.TextChoices):
        COMPUTER = "computer"
        NOTION = "notion"
        GITHUB = "github"

    user = models.ForeignKey(KhojUser, on_delete=models.CASCADE, default=None, null=True, blank=True)
    agent = models.ForeignKey(Agent, on_delete=models.CASCADE, default=None, null=True, blank=True)
    embeddings = VectorField(dimensions=None)
    raw = models.TextField()
    compiled = models.TextField()
    heading = models.CharField(max_length=1000, default=None, null=True, blank=True)
    file_source = models.CharField(max_length=30, choices=EntrySource.choices, default=EntrySource.COMPUTER)
    file_type = models.CharField(max_length=30, choices=EntryType.choices, default=EntryType.PLAINTEXT)
    file_path = models.CharField(max_length=400, default=None, null=True, blank=True)
    file_name = models.CharField(max_length=400, default=None, null=True, blank=True)
    url = models.URLField(max_length=400, default=None, null=True, blank=True)
    hashed_value = models.CharField(max_length=100)
    corpus_id = models.UUIDField(default=uuid.uuid4, editable=False)
    search_model = models.ForeignKey(SearchModelConfig, on_delete=models.SET_NULL, default=None, null=True, blank=True)
    file_object = models.ForeignKey(FileObject, on_delete=models.CASCADE, default=None, null=True, blank=True)

    def save(self, *args, **kwargs):
        if self.user and self.agent:
            raise ValidationError("An Entry cannot be associated with both a user and an agent.")


class EntryDates(DbBaseModel):
    date = models.DateField()
    entry = models.ForeignKey(Entry, on_delete=models.CASCADE, related_name="embeddings_dates")

    class Meta:
        indexes = [
            models.Index(fields=["date"]),
        ]


class UserRequests(DbBaseModel):
    """Stores user requests to the server for rate limiting."""

    user = models.ForeignKey(KhojUser, on_delete=models.CASCADE)
    slug = models.CharField(max_length=200)


class RateLimitRecord(DbBaseModel):
    """Stores individual request timestamps for rate limiting."""

    identifier = models.CharField(max_length=255, db_index=True)  # IP address or email
    slug = models.CharField(max_length=255, db_index=True)  # Differentiates limit types

    class Meta:
        indexes = [
            models.Index(fields=["identifier", "slug", "created_at"]),
        ]
        ordering = ["-created_at"]

    def __str__(self):
        return f"{self.slug} - {self.identifier} at {self.created_at}"


class DataStore(DbBaseModel):
    key = models.CharField(max_length=200, unique=True)
    value = models.JSONField(default=dict)
    private = models.BooleanField(default=False)
    owner = models.ForeignKey(KhojUser, on_delete=models.CASCADE, default=None, null=True, blank=True)


class McpServer(DbBaseModel):
    name = models.CharField(max_length=50, unique=True)
    path = models.CharField(max_length=200, unique=True)
    api_key = models.CharField(max_length=4000, blank=True, null=True)

    def __str__(self):
        return self.name


class UserMemory(DbBaseModel):
    """
    Long term memory store derived from conversation between user and agent.
    """

    user = models.ForeignKey(KhojUser, on_delete=models.CASCADE)
    agent = models.ForeignKey(Agent, on_delete=models.CASCADE, default=None, null=True, blank=True)
    embeddings = VectorField(dimensions=None)
    raw = models.TextField()
    search_model = models.ForeignKey(SearchModelConfig, on_delete=models.SET_NULL, default=None, null=True, blank=True)



# Generated by Django 5.1.10 on 2025-08-29 22:57

import django.db.models.deletion
import pgvector.django
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0098_alter_texttoimagemodelconfig_model_type"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserMemory",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("embeddings", pgvector.django.VectorField()),
                ("raw", models.TextField()),
                (
                    "agent",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="database.agent",
                    ),
                ),
                (
                    "search_model",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="database.searchmodelconfig",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="userconversationconfig",
            name="enable_memory",
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name="serverchatsettings",
            name="memory_mode",
            field=models.CharField(
                choices=[
                    ("disabled", "Disabled"),
                    ("enabled_default_off", "Enabled, default off"),
                    ("enabled_default_on", "Enabled, default on"),
                ],
                default="enabled_default_on",
                help_text="Server-level memory feature configuration. Disabled overrides user preference.",
                max_length=20,
            ),
        ),
    ]



import asyncio
import base64
import json
import logging
import os
from concurrent.futures import ThreadPoolExecutor

import requests
from fastapi import APIRouter, BackgroundTasks, Request, Response
from starlette.authentication import requires
from starlette.responses import RedirectResponse

from khoj.database.models import KhojUser, NotionConfig
from khoj.routers.helpers import configure_content
from khoj.utils.state import SearchType

NOTION_OAUTH_CLIENT_ID = os.getenv("NOTION_OAUTH_CLIENT_ID")
NOTION_OAUTH_CLIENT_SECRET = os.getenv("NOTION_OAUTH_CLIENT_SECRET")
NOTION_REDIRECT_URI = os.getenv("NOTION_REDIRECT_URI")

notion_router = APIRouter()

executor = ThreadPoolExecutor()

logger = logging.getLogger(__name__)


async def run_in_executor(func, *args):
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(executor, func, *args)


@notion_router.get("/auth/callback")
@requires(["authenticated"], redirect="login_page")
async def notion_auth_callback(request: Request, background_tasks: BackgroundTasks):
    code = request.query_params.get("code")
    state = request.query_params.get("state")
    if not code or not state:
        return Response("Missing code or state", status_code=400)

    # Use authenticated user from session instead of trusting the state parameter
    user: KhojUser = request.user.object

    # Verify state matches user UUID as CSRF protection
    if state != str(user.uuid):
        logger.warning(f"Notion OAuth state mismatch for user {user.uuid}")
        return Response("Invalid state parameter", status_code=400)

    await NotionConfig.objects.filter(user=user).adelete()

    bearer_token = f"{NOTION_OAUTH_CLIENT_ID}:{NOTION_OAUTH_CLIENT_SECRET}"
    base64_encoded_token = base64.b64encode(bearer_token.encode()).decode()

    headers = {
        "Accept": "application/json",
        "Content-Type": "application/json",
        "Authorization": f"Basic {base64_encoded_token}",
    }

    data = {
        "grant_type": "authorization_code",
        "code": code,
        "redirect_uri": NOTION_REDIRECT_URI,
    }

    response = requests.post("https://api.notion.com/v1/oauth/token", data=json.dumps(data), headers=headers)

    final_response = response.json()

    logger.info(f"Notion auth callback response: {final_response}")

    access_token = final_response.get("access_token")
    await NotionConfig.objects.acreate(token=access_token, user=user)

    owner = final_response.get("owner")
    workspace_id = final_response.get("workspace_id")
    workspace_name = final_response.get("workspace_name")
    bot_id = final_response.get("bot_id")

    logger.info(
        f"Notion integration. Owner: {owner}, Workspace ID: {workspace_id}, Workspace Name: {workspace_name}, Bot ID: {bot_id}"
    )

    notion_redirect = str(request.app.url_path_for("config_page"))

    # Trigger an async job to configure_content. Let it run without blocking the response.
    background_tasks.add_task(run_in_executor, configure_content, user, {}, False, SearchType.Notion)

    return RedirectResponse(notion_redirect)



import json
import logging
from datetime import datetime
from textwrap import dedent
from typing import List, Optional

from khoj.database.models import ChatMessageModel, ChatModel
from khoj.processor.conversation.utils import (
    AgentMessage,
    OperatorRun,
    construct_structured_message,
)
from khoj.processor.operator.grounding_agent import GroundingAgent
from khoj.processor.operator.grounding_agent_uitars import GroundingAgentUitars
from khoj.processor.operator.operator_actions import OperatorAction, WaitAction
from khoj.processor.operator.operator_agent_base import AgentActResult, OperatorAgent
from khoj.processor.operator.operator_environment_base import (
    EnvironmentType,
    EnvState,
    EnvStepResult,
)
from khoj.routers.helpers import send_message_to_model_wrapper
from khoj.utils.helpers import get_openai_async_client, is_none_or_empty

logger = logging.getLogger(__name__)


# --- Binary Operator Agent ---
class BinaryOperatorAgent(OperatorAgent):
    """
    An OperatorAgent that uses two LLMs:
    1. Reasoning LLM: Determines the next high-level action based on the objective and current visual reasoning trajectory.
    2. Grounding LLM: Converts the high-level action into specific, actions executable on the environment.
    """

    def __init__(
        self,
        query: str,
        reasoning_model: ChatModel,
        grounding_model: ChatModel,
        environment_type: EnvironmentType,
        max_iterations: int,
        max_context: int,
        chat_history: List[AgentMessage] = [],
        previous_trajectory: Optional[OperatorRun] = None,
        tracer: dict = {},
    ):
        super().__init__(
            query,
            reasoning_model,
            environment_type,
            max_iterations,
            max_context,
            chat_history,
            previous_trajectory,
            tracer,
        )  # Use reasoning model for primary tracking
        self.reasoning_model = reasoning_model
        self.grounding_model = grounding_model
        # Initialize openai api compatible client for grounding model
        grounding_client = get_openai_async_client(
            grounding_model.ai_model_api.api_key, grounding_model.ai_model_api.api_base_url
        )

        self.grounding_agent: GroundingAgent | GroundingAgentUitars = None
        if "ui-tars-1.5" in grounding_model.name:
            self.grounding_agent = GroundingAgentUitars(
                grounding_model.name, self.environment_type, grounding_client, max_iterations, tracer=tracer
            )
        else:
            self.grounding_agent = GroundingAgent(
                grounding_model.name, self.environment_type, grounding_client, max_iterations, tracer=tracer
            )

    async def act(self, current_state: EnvState) -> AgentActResult:
        """
        Uses a two-step LLM process to determine and structure the next action.
        """
        self._commit_trace()  # Commit trace before next action

        # --- Step 1: Reasoning LLM determines high-level action ---
        reasoner_response = await self.act_reason(current_state)
        natural_language_action = reasoner_response["message"]
        if reasoner_response["type"] == "error":
            logger.error(natural_language_action)
            return AgentActResult(
                actions=[],
                action_results=[],
                rendered_response={"text": natural_language_action, "image": None},
            )
        elif reasoner_response["type"] == "done":
            return AgentActResult(
                actions=[],
                action_results=[],
                rendered_response={"text": natural_language_action, "image": None},
            )

        # --- Step 2: Grounding LLM converts NL action to structured action ---
        return await self.act_ground(natural_language_action, current_state)

    async def act_reason(self, current_state: EnvState) -> dict[str, str]:
        """
        Uses the reasoning LLM to determine the next high-level action based on the operation trajectory.
        """
        reasoning_system_prompt = self.get_instruction(self.environment_type, current_state)
        if is_none_or_empty(self.messages):
            query_text = f"**Main Objective**: {self.query}"
            query_screenshot = [f"data:image/webp;base64,{current_state.screenshot}"]
            first_message_content = construct_structured_message(
                message=query_text,
                images=query_screenshot,
                model_type=self.reasoning_model.model_type,
                vision_enabled=True,
            )
            current_message = AgentMessage(role="user", content=first_message_content)
        else:
            current_message = self.messages.pop()
            query_text = self._get_message_text(current_message)
            query_screenshot = self._get_message_images(current_message)

        # Construct input for visual reasoner history
        visual_reasoner_history = self._format_message_for_api(self.messages)
        try:
            raw_response = await send_message_to_model_wrapper(
                query=query_text,
                query_images=query_screenshot,
                system_message=reasoning_system_prompt,
                chat_history=visual_reasoner_history,
                agent_chat_model=self.reasoning_model,
                tracer=self.tracer,
            )
            natural_language_action = raw_response.text

            if not isinstance(natural_language_action, str) or not natural_language_action.strip():
                raise ValueError(f"Natural language action is empty or not a string. Got {natural_language_action}")

            self.messages.append(current_message)
            self.messages.append(AgentMessage(role="assistant", content=natural_language_action))

            if natural_language_action.strip().endswith("DONE"):
                return {"type": "done", "message": "Completed task."}

            logger.info(f"Reasoning LLM suggested action: {natural_language_action}")

        except Exception as e:
            logger.error(f"Error calling Reasoning LLM: {e}", exc_info=True)
            return {"type": "error", "message": f"Error calling Reasoning LLM: {e}"}

        return {"type": "action", "message": natural_language_action}

    async def act_ground(self, action_instruction: str, current_state: EnvState) -> AgentActResult:
        """Uses the grounding LLM to convert the high-level action into structured browser actions."""
        actions: List[OperatorAction] = []
        action_results: List[dict] = []
        rendered_parts = [f"**Thought (Vision)**: {action_instruction}"]

        try:
            grounding_response, actions = await self.grounding_agent.act(action_instruction, current_state)

            # Process grounding response
            if grounding_response.strip().endswith("DONE"):
                # Ignore DONE response by the grounding agent. Reasoning agent handles termination.
                actions.append(WaitAction(duration=1.0))
                rendered_parts += ["Nothing to do."]
            elif grounding_response.strip().endswith("FAIL"):
                # Ignore FAIL response by the grounding agent. Reasoning agent handles termination.
                actions.append(WaitAction(duration=1.0))
                rendered_parts += ["Could not process response."]
            else:
                grounding_thoughts = grounding_response.rsplit("\nAction: ", 1)[0]
                rendered_parts += [f"**Thought (Grounding)**: {grounding_thoughts}"]
                for action in actions:
                    if action.type == "type":
                        rendered_parts += [f'**Action**: Type "{action.text}"']
                    elif action.type == "keypress":
                        rendered_parts += [f'**Action**: Press "{action.keys}"']
                    elif action.type == "hold_key":
                        rendered_parts += [f'**Action**: Hold "{action.text}" for {action.duration} seconds']
                    elif action.type == "key_up":
                        rendered_parts += [f'**Action**: Release Key "{action.key}"']
                    elif action.type == "key_down":
                        rendered_parts += [f'**Action**: Press Key "{action.key}"']
                    elif action.type == "screenshot" and not current_state.screenshot:
                        rendered_parts += ["**Error**: Failed to take screenshot"]
                    elif action.type == "goto":
                        rendered_parts += [f"**Action**: Open URL {action.url}"]
                    else:
                        rendered_parts += [f"**Action**: {action.type}"]
            action_results += [{"content": None}]  # content set after environment step
        except Exception as e:
            logger.error(f"Error calling Grounding LLM: {e}", exc_info=True)
            rendered_parts += [f"**Error**: Error contacting Grounding LLM: {e}"]

        rendered_response = self._render_response(rendered_parts, current_state.screenshot)

        return AgentActResult(
            actions=actions,
            action_results=action_results,
            rendered_response=rendered_response,
        )

    def add_action_results(self, env_steps: list[EnvStepResult], agent_action: AgentActResult) -> None:
        """
        Adds the results of executed actions back into the message history,
        formatted for the next OpenAI vision LLM call.
        """
        if not agent_action.action_results:
            return

        for idx, env_step in enumerate(env_steps):
            result_content = env_step.error or env_step.output or "[Action completed]"
            action_result = agent_action.action_results[idx]
            if env_step.type == "image":
                message = "**Action Result**: Took screenshot"
                images = [f"data:image/png;base64,{env_step.screenshot_base64}"]
            elif idx == len(env_steps) - 1:
                message = f"**Action Result**: {json.dumps(result_content)}"
                images = [f"data:image/png;base64,{env_step.screenshot_base64}"]
            else:
                message = f"**Action Result**: {json.dumps(result_content)}"
                images = []
            if not images:
                action_result["content"] = [{"type": "text", "text": message}]
            else:
                action_result["content"] = construct_structured_message(
                    message=message,
                    images=images,
                    model_type=self.reasoning_model.model_type,
                    vision_enabled=True,
                )

        # Append action results to history
        action_results_content = []
        for action_result in agent_action.action_results:
            if not action_result.get("content"):
                logger.error("Action result content is empty or None: {action_result}")
            action_results_content.extend(action_result["content"])
        self.messages.append(AgentMessage(role="environment", content=action_results_content))

    async def summarize(self, env_state: EnvState, summarize_prompt: str = None) -> str:
        summarize_prompt = summarize_prompt or self.summarize_prompt
        conversation_history = self._format_message_for_api(self.messages)
        try:
            summary = await send_message_to_model_wrapper(
                query=summarize_prompt,
                chat_history=conversation_history,
                agent_chat_model=self.reasoning_model,
                tracer=self.tracer,
            )
            # Set summary to last action message
            if not summary:
                raise ValueError("Summary is empty.")
        except Exception as e:
            logger.error(f"Error calling Reasoning LLM for summary: {e}")
            summary = "\n".join([self._get_message_text(msg) for msg in self.messages])

        # Append summary messages to history
        trigger_summary = AgentMessage(role="user", content=summarize_prompt)
        summary_message = AgentMessage(role="assistant", content=summary.text)
        self.messages.extend([trigger_summary, summary_message])

        return summary.text

    def _compile_response(self, response_content: str | List) -> str:
        """Compile response content into a string, handling OpenAI message structures."""
        if isinstance(response_content, str):
            return response_content

        if isinstance(response_content, list):  # Tool results list
            compiled = ["**Tool Results**:"]
            for item in response_content:
                if isinstance(item, dict) and item.get("role") == "tool":
                    compiled.append(f"  - ID {item.get('tool_call_id')}: {item.get('content')}")
                else:
                    compiled.append(f"  - {str(item)}")  # Fallback
            return "\n".join(compiled)

        # Fallback for unexpected types
        return str(response_content)

    def _render_response(self, response: List, screenshot: str | None) -> dict:
        """Render response for display"""
        render_payload = {
            "text": "\n- ".join(response),
            "image": f"data:image/webp;base64,{screenshot}" if screenshot else None,
        }
        return render_payload

    def _get_message_text(self, message: AgentMessage) -> str:
        if isinstance(message.content, list):
            return "\n".join([item["text"] for item in message.content if item["type"] == "text"])
        return message.content

    def _get_message_images(self, message: AgentMessage) -> List[str]:
        images = []
        if isinstance(message.content, list):
            images = [item["image_url"]["url"] for item in message.content if item["type"] == "image_url"]
        return images

    def _format_message_for_api(self, messages: list[AgentMessage]) -> List[ChatMessageModel]:
        """Format operator agent messages into the Khoj conversation history format."""
        formatted_messages = [
            ChatMessageModel(
                message=self._get_message_text(message),
                images=self._get_message_images(message),
                by="you" if message.role in ["user", "environment"] else message.role,
            )
            for message in messages
        ]
        return formatted_messages

    def get_instruction(self, environment_type: EnvironmentType, env_state: EnvState) -> str:
        """Get the system instruction for the reasoning agent."""
        if environment_type == EnvironmentType.BROWSER:
            return dedent(
                f"""
                # Introduction
                * You are Khoj, a smart and resourceful web browsing assistant. You help the user accomplish their task using a web browser.
                * You are given the user's query and screenshots of the browser's state transitions.
                * The current date is {datetime.today().strftime("%A, %B %-d, %Y")}.
                * The current URL is {env_state.url}.

                # Your Task
                * First look at the screenshots carefully to notice all pertinent information.
                * Then instruct a tool AI to perform the next action that will help you progress towards the user's goal.
                * Make sure you scroll down to see everything before deciding something isn't available.
                * Perform web searches using DuckDuckGo. Don't use Google even if requested as the query will fail.
                * Use your creativity to find alternate ways to make progress if you get stuck at any point.

                # Tool AI Capabilities
                * The tool AI only has access to the current screenshot and your instructions. It uses your instructions to perform the next action on the page.
                * It can interact with the web browser with these actions: click, right click, double click, type, scroll, drag, wait, goto url and go back to previous page.
                * It cannot access the OS, filesystem or application window. It just controls a single Chromium browser tab via Playwright.

                # IMPORTANT
                * You are allowed upto {self.max_iterations} iterations to complete the task.
                * To navigate to a specific URL, put "GOTO <URL>" (without quotes) on the last line of your response.
                * To navigate back to the previous page, end your response with "BACK" (without quotes).
                * Once you've verified that the main objective has been achieved, end your response with "DONE" (without quotes).

                # Examples
                ## Example 1
                GOTO https://example.com
                ## Example 2
                click the blue login button located at the top right corner
                ## Example 3
                scroll down the page
                ## Example 4
                type the username example@email.com into the input field labeled Username
                ## Example 5
                DONE

                # Instructions
                Now describe a single high-level action to take next to progress towards the user's goal in detail.
                Focus on the visual action and provide all necessary context.
                """
            ).strip()

        elif environment_type == EnvironmentType.COMPUTER:
            return dedent(
                f"""
                # Introduction
                * You are Khoj, a smart and resourceful computer assistant. You help the user accomplish their task using a computer.
                * You are given the user's query and screenshots of the computer's state transitions.
                * The current date is {datetime.today().strftime("%A, %B %-d, %Y")}.

                # Your Task
                * First look at the screenshots carefully to notice all pertinent information.
                * Then instruct a tool AI to perform the next action that will help you progress towards the user's goal.
                * Make sure you scroll down to see everything before deciding something isn't available.
                * Perform web searches using DuckDuckGo. Don't use Google even if requested as the query will fail.
                * Use your creativity to find alternate ways to make progress if you get stuck at any point.

                # Tool AI Capabilities
                * The tool AI only has access to the current screenshot and your instructions. It uses your instructions to perform the next action on the page.
                * It can interact with the computer with these actions: click, right click, double click, type, scroll, drag, wait to previous page.

                # IMPORTANT
                * You are allowed upto {self.max_iterations} iterations to complete the task.
                * Once you've verified that the main objective has been achieved, end your response with "DONE" (without quotes).

                # Examples
                ## Example 1
                type https://example.com into the address bar and press Enter
                ## Example 2
                click the blue login button located at the top right corner
                ## Example 3
                scroll down the page
                ## Example 4
                type the username example@email.com into the input field labeled Username
                ## Example 5
                DONE

                # Instructions
                Now describe a single high-level action to take next to progress towards the user's goal in detail.
                Focus on the visual action and provide all necessary context.
                """
            ).strip()
        else:
            raise ValueError(f"Expected environment type: Computer or Browser. Got {environment_type}.")

    def reset(self):
        """Reset the agent state."""
        super().reset()
        self.grounding_agent.reset()  # Reset grounding agent state



import asyncio
import base64
import json
import logging
import time
import uuid
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from typing import Any, Dict, List, Optional
from urllib.parse import unquote

from fastapi import (
    APIRouter,
    Depends,
    HTTPException,
    Request,
    WebSocket,
    WebSocketDisconnect,
)
from fastapi.responses import RedirectResponse, Response, StreamingResponse
from fastapi.websockets import WebSocketState
from starlette.authentication import has_required_scope, requires
from starlette.requests import URL, Headers

from khoj.app.settings import ALLOWED_HOSTS
from khoj.database.adapters import (
    AgentAdapters,
    ConversationAdapters,
    EntryAdapters,
    PublicConversationAdapters,
    UserMemoryAdapters,
    aget_user_name,
)
from khoj.database.models import Agent, KhojUser
from khoj.processor.conversation import prompts
from khoj.processor.conversation.openai.utils import is_local_api
from khoj.processor.conversation.prompts import no_entries_found
from khoj.processor.conversation.utils import (
    OperatorRun,
    ResponseWithThought,
    defilter_query,
    save_to_conversation_log,
)
from khoj.processor.image.generate import text_to_image
from khoj.processor.operator import operate_environment
from khoj.processor.speech.text_to_speech import TextToSpeechError, generate_text_to_speech
from khoj.processor.tools.online_search import (
    deduplicate_organic_results,
    read_webpages,
    search_online,
)
from khoj.processor.tools.run_code import run_code
from khoj.routers.email import send_query_feedback
from khoj.routers.helpers import (
    ApiImageRateLimiter,
    ApiUserRateLimiter,
    ChatEvent,
    ChatRequestBody,
    CommonQueryParams,
    ConversationCommandRateLimiter,
    DeleteMessageRequestBody,
    FeedbackData,
    WebSocketConnectionManager,
    acreate_title_from_history,
    agenerate_chat_response,
    aget_data_sources_and_output_format,
    gather_raw_query_files,
    generate_mermaidjs_diagram,
    get_conversation_command,
    get_message_from_queue,
    is_query_empty,
    is_ready_to_chat,
    read_chat_stream,
    search_documents,
    update_telemetry_state,
    validate_chat_model,
)
from khoj.routers.research import ResearchIteration, research
from khoj.routers.storage import upload_user_image_to_bucket
from khoj.utils import state
from khoj.utils.helpers import (
    ConversationCommand,
    clean_text_for_db,
    command_descriptions,
    convert_image_to_webp,
    get_country_code_from_timezone,
    get_country_name_from_timezone,
    is_env_var_true,
    is_none_or_empty,
    is_operator_enabled,
)
from khoj.utils.rawconfig import (
    FileFilterRequest,
    FilesFilterRequest,
    LocationData,
)

# Initialize Router
logger = logging.getLogger(__name__)
conversation_command_rate_limiter = ConversationCommandRateLimiter(
    trial_rate_limit=20, subscribed_rate_limit=75, slug="command"
)

api_chat = APIRouter()


@api_chat.get("/stats", response_class=Response)
@requires(["authenticated"])
def chat_stats(request: Request, common: CommonQueryParams) -> Response:
    num_conversations = ConversationAdapters.get_num_conversations(request.user.object)
    return Response(
        content=json.dumps({"num_conversations": num_conversations}), media_type="application/json", status_code=200
    )


@api_chat.get("/export", response_class=Response)
@requires(["authenticated"])
def export_conversation(request: Request, common: CommonQueryParams, page: Optional[int] = 1) -> Response:
    all_conversations = ConversationAdapters.get_all_conversations_for_export(request.user.object, page=page)
    return Response(content=json.dumps(all_conversations), media_type="application/json", status_code=200)


@api_chat.get("/conversation/file-filters/{conversation_id}", response_class=Response)
@requires(["authenticated"])
def get_file_filter(request: Request, conversation_id: str) -> Response:
    conversation = ConversationAdapters.get_conversation_by_user(request.user.object, conversation_id=conversation_id)
    if not conversation:
        return Response(content=json.dumps({"status": "error", "message": "Conversation not found"}), status_code=404)

    # get all files from "computer"
    file_list = EntryAdapters.get_all_filenames_by_source(request.user.object, "computer")
    file_filters = []
    for file in conversation.file_filters:
        if file in file_list:
            file_filters.append(file)
    return Response(content=json.dumps(file_filters), media_type="application/json", status_code=200)


@api_chat.delete("/conversation/file-filters/bulk", response_class=Response)
@requires(["authenticated"])
def remove_files_filter(request: Request, filter: FilesFilterRequest) -> Response:
    conversation_id = filter.conversation_id
    files_filter = filter.filenames
    file_filters = ConversationAdapters.remove_files_from_filter(request.user.object, conversation_id, files_filter)
    return Response(content=json.dumps(file_filters), media_type="application/json", status_code=200)


@api_chat.post("/conversation/file-filters/bulk", response_class=Response)
@requires(["authenticated"])
def add_files_filter(request: Request, filter: FilesFilterRequest):
    try:
        conversation_id = filter.conversation_id
        files_filter = filter.filenames
        file_filters = ConversationAdapters.add_files_to_filter(request.user.object, conversation_id, files_filter)
        return Response(content=json.dumps(file_filters), media_type="application/json", status_code=200)
    except Exception as e:
        logger.error(f"Error adding file filter {filter.filenames}: {e}", exc_info=True)
        raise HTTPException(status_code=422, detail=str(e))


@api_chat.post("/conversation/file-filters", response_class=Response)
@requires(["authenticated"])
def add_file_filter(request: Request, filter: FileFilterRequest):
    try:
        conversation_id = filter.conversation_id
        files_filter = [filter.filename]
        file_filters = ConversationAdapters.add_files_to_filter(request.user.object, conversation_id, files_filter)
        return Response(content=json.dumps(file_filters), media_type="application/json", status_code=200)
    except Exception as e:
        logger.error(f"Error adding file filter {filter.filename}: {e}", exc_info=True)
        raise HTTPException(status_code=422, detail=str(e))


@api_chat.delete("/conversation/file-filters", response_class=Response)
@requires(["authenticated"])
def remove_file_filter(request: Request, filter: FileFilterRequest) -> Response:
    conversation_id = filter.conversation_id
    files_filter = [filter.filename]
    file_filters = ConversationAdapters.remove_files_from_filter(request.user.object, conversation_id, files_filter)
    return Response(content=json.dumps(file_filters), media_type="application/json", status_code=200)


@api_chat.post("/feedback")
@requires(["authenticated"])
async def sendfeedback(request: Request, data: FeedbackData):
    user: KhojUser = request.user.object
    await send_query_feedback(data.uquery, data.kquery, data.sentiment, user.email)


@api_chat.post("/speech")
@requires(["authenticated"])
async def text_to_speech(
    request: Request,
    common: CommonQueryParams,
    text: str,
    rate_limiter_per_minute=Depends(
        ApiUserRateLimiter(requests=30, subscribed_requests=30, window=60, slug="chat_minute")
    ),
    rate_limiter_per_day=Depends(
        ApiUserRateLimiter(requests=100, subscribed_requests=600, window=60 * 60 * 24, slug="chat_day")
    ),
) -> Response:
    voice_model = await ConversationAdapters.aget_voice_model_config(request.user.object)

    params = {"text_to_speak": text}

    if voice_model:
        params["voice_id"] = voice_model.model_id

    try:
        speech_stream = generate_text_to_speech(**params)
    except TextToSpeechError as e:
        raise HTTPException(status_code=503, detail=str(e))
    return StreamingResponse(speech_stream.iter_content(chunk_size=1024), media_type="audio/mpeg")


@api_chat.get("/starters", response_class=Response)
@requires(["authenticated"])
async def chat_starters(
    request: Request,
    common: CommonQueryParams,
) -> Response:
    user: KhojUser = request.user.object
    starter_questions = await ConversationAdapters.aget_conversation_starters(user)
    return Response(content=json.dumps(starter_questions), media_type="application/json", status_code=200)


@api_chat.get("/history")
@requires(["authenticated"])
def chat_history(
    request: Request,
    common: CommonQueryParams,
    conversation_id: Optional[str] = None,
    n: Optional[int] = None,
):
    user = request.user.object
    validate_chat_model(user)

    # Load Conversation History
    conversation = ConversationAdapters.get_conversation_by_user(
        user=user, client_application=request.user.client_app, conversation_id=conversation_id
    )

    if conversation is None:
        return Response(
            content=json.dumps({"status": "error", "message": f"Conversation: {conversation_id} not found"}),
            status_code=404,
        )

    agent_metadata = None
    if conversation.agent:
        if conversation.agent.privacy_level == Agent.PrivacyLevel.PRIVATE and conversation.agent.creator != user:
            conversation.agent = None
        else:
            agent_metadata = {
                "slug": conversation.agent.slug,
                "name": conversation.agent.name,
                "is_creator": conversation.agent.creator == user,
                "color": conversation.agent.style_color,
                "icon": conversation.agent.style_icon,
                "persona": conversation.agent.personality,
                "is_hidden": conversation.agent.is_hidden,
            }

    meta_log = conversation.conversation_log
    meta_log.update(
        {
            "conversation_id": conversation.id,
            "slug": conversation.title if conversation.title else conversation.slug,
            "agent": agent_metadata,
            "is_owner": conversation.user == user,
        }
    )

    if n:
        # Get latest N messages if N > 0
        if n > 0 and meta_log.get("chat"):
            meta_log["chat"] = meta_log["chat"][-n:]
        # Else return all messages except latest N
        elif n < 0 and meta_log.get("chat"):
            meta_log["chat"] = meta_log["chat"][:n]

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="chat_history",
        **common.__dict__,
    )

    return {"status": "ok", "response": meta_log}


@api_chat.get("/share/history")
def get_shared_chat(
    request: Request,
    common: CommonQueryParams,
    public_conversation_slug: str,
    n: Optional[int] = None,
):
    user = request.user.object if request.user.is_authenticated else None

    # Load Conversation History
    conversation = PublicConversationAdapters.get_public_conversation_by_slug(public_conversation_slug)

    if conversation is None:
        return Response(
            content=json.dumps({"status": "error", "message": f"Conversation: {public_conversation_slug} not found"}),
            status_code=404,
        )

    agent_metadata = None
    if conversation.agent:
        if conversation.agent.privacy_level == Agent.PrivacyLevel.PRIVATE and conversation.agent.creator != user:
            if conversation.agent.is_hidden:
                default_agent = AgentAdapters.get_default_agent()
                agent_metadata = {
                    "slug": default_agent.slug,
                    "name": default_agent.name,
                    "is_creator": False,
                    "color": default_agent.style_color,
                    "icon": default_agent.style_icon,
                    "persona": default_agent.personality,
                    "is_hidden": default_agent.is_hidden,
                }
            else:
                conversation.agent = None
        else:
            agent_metadata = {
                "slug": conversation.agent.slug,
                "name": conversation.agent.name,
                "is_creator": conversation.agent.creator == user,
                "color": conversation.agent.style_color,
                "icon": conversation.agent.style_icon,
                "persona": conversation.agent.personality,
                "is_hidden": conversation.agent.is_hidden,
            }

    meta_log = conversation.conversation_log
    scrubbed_title = conversation.title if conversation.title else conversation.slug

    if scrubbed_title:
        scrubbed_title = scrubbed_title.replace("-", " ")

    meta_log.update(
        {
            "conversation_id": conversation.id,
            "slug": scrubbed_title,
            "agent": agent_metadata,
            "is_owner": conversation.source_owner == user,
        }
    )

    if n:
        # Get latest N messages if N > 0
        if n > 0 and meta_log.get("chat"):
            meta_log["chat"] = meta_log["chat"][-n:]
        # Else return all messages except latest N
        elif n < 0 and meta_log.get("chat"):
            meta_log["chat"] = meta_log["chat"][:n]

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="get_shared_chat_history",
        **common.__dict__,
    )

    return {"status": "ok", "response": meta_log}


@api_chat.delete("/history")
@requires(["authenticated"])
async def clear_chat_history(
    request: Request,
    common: CommonQueryParams,
    conversation_id: Optional[str] = None,
):
    user = request.user.object

    # Clear Conversation History
    await ConversationAdapters.adelete_conversation_by_user(user, request.user.client_app, conversation_id)

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="clear_chat_history",
        **common.__dict__,
    )

    return {"status": "ok", "message": "Conversation history cleared"}


@api_chat.post("/share/fork")
@requires(["authenticated"])
def fork_public_conversation(
    request: Request,
    common: CommonQueryParams,
    public_conversation_slug: str,
):
    user = request.user.object

    # Load Conversation History
    public_conversation = PublicConversationAdapters.get_public_conversation_by_slug(public_conversation_slug)

    # Duplicate Public Conversation to User's Private Conversation
    new_conversation = ConversationAdapters.create_conversation_from_public_conversation(
        user, public_conversation, request.user.client_app
    )

    chat_metadata = {"forked_conversation": public_conversation.slug}

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="fork_public_conversation",
        **common.__dict__,
        metadata=chat_metadata,
    )

    redirect_uri = str(request.app.url_path_for("chat_page"))

    return Response(
        status_code=200,
        content=json.dumps(
            {
                "status": "ok",
                "next_url": redirect_uri,
                "conversation_id": str(new_conversation.id),
            }
        ),
    )


@api_chat.post("/share")
@requires(["authenticated"])
def duplicate_chat_history_public_conversation(
    request: Request,
    common: CommonQueryParams,
    conversation_id: str,
):
    user = request.user.object
    domain = request.headers.get("host")
    scheme = request.url.scheme
    # Force https upgrade if not explicitly disabled and not local host
    if scheme == "http" and not is_env_var_true("KHOJ_NO_HTTPS") and not is_local_api(f"{request.base_url}"):
        scheme = "https"

    # Throw unauthorized exception if domain not in ALLOWED_HOSTS
    host_domain = domain.split(":")[0]
    if host_domain not in ALLOWED_HOSTS:
        raise HTTPException(status_code=401, detail="Unauthorized domain")

    # Duplicate Conversation History to Public Conversation
    conversation = ConversationAdapters.get_conversation_by_user(user, request.user.client_app, conversation_id)
    public_conversation = ConversationAdapters.make_public_conversation_copy(conversation)
    public_conversation_url = PublicConversationAdapters.get_public_conversation_url(public_conversation)

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="post_chat_share",
        **common.__dict__,
    )

    return Response(
        status_code=200, content=json.dumps({"status": "ok", "url": f"{scheme}://{domain}{public_conversation_url}"})
    )


@api_chat.delete("/share")
@requires(["authenticated"])
def delete_public_conversation(
    request: Request,
    common: CommonQueryParams,
    public_conversation_slug: str,
):
    user = request.user.object

    # Delete Public Conversation
    PublicConversationAdapters.delete_public_conversation_by_slug(user=user, slug=public_conversation_slug)

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="delete_chat_share",
        **common.__dict__,
    )

    # Redirect to the main chat page
    redirect_uri = str(request.app.url_path_for("chat_page"))
    return RedirectResponse(
        url=redirect_uri,
        status_code=301,
    )


@api_chat.get("/sessions")
@requires(["authenticated"])
def chat_sessions(
    request: Request,
    common: CommonQueryParams,
    recent: Optional[bool] = False,
):
    user = request.user.object

    # Load Conversation Sessions
    conversations = ConversationAdapters.get_conversation_sessions(user, request.user.client_app)
    if recent:
        conversations = conversations[:8]

    sessions = conversations.values_list(
        "id",
        "slug",
        "title",
        "agent__slug",
        "agent__name",
        "created_at",
        "updated_at",
        "agent__style_icon",
        "agent__style_color",
        "agent__is_hidden",
    )

    session_values = [
        {
            "conversation_id": str(session[0]),
            "slug": session[2] or session[1],
            "agent_name": session[4],
            "created": session[5].strftime("%Y-%m-%d %H:%M:%S"),
            "updated": session[6].strftime("%Y-%m-%d %H:%M:%S"),
            "agent_icon": session[7],
            "agent_color": session[8],
            "agent_is_hidden": session[9],
        }
        for session in sessions
    ]

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="chat_sessions",
        **common.__dict__,
    )

    return Response(content=json.dumps(session_values), media_type="application/json", status_code=200)


@api_chat.post("/sessions")
@requires(["authenticated"])
async def create_chat_session(
    request: Request,
    common: CommonQueryParams,
    agent_slug: Optional[str] = None,
    # Add parameters here to create a custom hidden agent on the fly
):
    user = request.user.object

    # Create new Conversation Session
    conversation = await ConversationAdapters.acreate_conversation_session(user, request.user.client_app, agent_slug)

    response = {"conversation_id": str(conversation.id)}

    conversation_metadata = {
        "agent": agent_slug,
    }

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="create_chat_sessions",
        metadata=conversation_metadata,
        **common.__dict__,
    )

    return Response(content=json.dumps(response), media_type="application/json", status_code=200)


@api_chat.get("/options", response_class=Response)
async def chat_options(
    request: Request,
    common: CommonQueryParams,
) -> Response:
    cmd_options = {}
    for cmd in ConversationCommand:
        if cmd == ConversationCommand.Operator and not is_operator_enabled():
            continue
        if cmd in command_descriptions:
            cmd_options[cmd.value] = command_descriptions[cmd]

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="chat_options",
        **common.__dict__,
    )
    return Response(content=json.dumps(cmd_options), media_type="application/json", status_code=200)


@api_chat.patch("/title", response_class=Response)
@requires(["authenticated"])
async def set_conversation_title(
    request: Request,
    common: CommonQueryParams,
    title: str,
    conversation_id: Optional[str] = None,
) -> Response:
    user = request.user.object
    title = title.strip()[:200]

    # Set Conversation Title
    conversation = await ConversationAdapters.aset_conversation_title(
        user, request.user.client_app, conversation_id, title
    )

    success = True if conversation else False

    update_telemetry_state(
        request=request,
        telemetry_type="api",
        api="set_conversation_title",
        **common.__dict__,
    )

    return Response(
        content=json.dumps({"status": "ok", "success": success}), media_type="application/json", status_code=200
    )


@api_chat.post("/title")
@requires(["authenticated"])
async def generate_chat_title(
    request: Request,
    common: CommonQueryParams,
    conversation_id: str,
):
    user: KhojUser = request.user.object
    conversation = await ConversationAdapters.aget_conversation_by_user(user=user, conversation_id=conversation_id)

    # Conversation.title is explicitly set by the user. Do not override.
    if conversation.title:
        return {"status": "ok", "title": conversation.title}

    if not conversation:
        raise HTTPException(status_code=404, detail="Conversation not found")

    new_title = await acreate_title_from_history(request.user.object, conversation=conversation)
    conversation.slug = clean_text_for_db(new_title[:200])

    await conversation.asave()

    return {"status": "ok", "title": new_title}


@api_chat.delete("/conversation/message", response_class=Response)
@requires(["authenticated"])
def delete_message(request: Request, delete_request: DeleteMessageRequestBody) -> Response:
    user = request.user.object
    success = ConversationAdapters.delete_message_by_turn_id(
        user, delete_request.conversation_id, delete_request.turn_id
    )
    if success:
        return Response(content=json.dumps({"status": "ok"}), media_type="application/json", status_code=200)
    else:
        return Response(content=json.dumps({"status": "error", "message": "Message not found"}), status_code=404)


async def event_generator(
    body: ChatRequestBody,
    user_scope: Any,
    common: CommonQueryParams,
    headers: Headers,
    request_obj: Request | WebSocket,
    parent_interrupt_queue: asyncio.Queue = None,
):
    # Access the parameters from the body
    q = body.q
    n = body.n
    d = body.d
    stream = body.stream
    title = body.title
    conversation_id = body.conversation_id
    turn_id = str(body.turn_id or uuid.uuid4())
    city = body.city
    region = body.region
    country = body.country or get_country_name_from_timezone(body.timezone)
    country_code = body.country_code or get_country_code_from_timezone(body.timezone)
    raw_images = body.images
    raw_query_files = body.files

    start_time = time.perf_counter()
    ttft = None
    chat_metadata: dict = {}
    conversation = None
    user: KhojUser = user_scope.object
    is_subscribed = has_required_scope(request_obj, ["premium"])
    q = unquote(q)
    defiltered_query = defilter_query(q)
    train_of_thought = []
    cancellation_event = asyncio.Event()
    child_interrupt_queue: asyncio.Queue = asyncio.Queue(maxsize=10)

    tracer: dict = {
        "mid": turn_id,
        "cid": conversation_id,
        "uid": user.id,
        "khoj_version": state.khoj_version,
    }

    uploaded_images: list[str] = []
    if raw_images:
        for image in raw_images:
            decoded_string = unquote(image)
            base64_data = decoded_string.split(",", 1)[1]
            image_bytes = base64.b64decode(base64_data)
            webp_image_bytes = convert_image_to_webp(image_bytes)
            uploaded_image = upload_user_image_to_bucket(webp_image_bytes, user.id)
            if not uploaded_image:
                base64_webp_image = base64.b64encode(webp_image_bytes).decode("utf-8")
                uploaded_image = f"data:image/webp;base64,{base64_webp_image}"
            uploaded_images.append(uploaded_image)

    query_files: Dict[str, str] = {}
    if raw_query_files:
        for file in raw_query_files:
            query_files[file.name] = file.content

    research_results: List[ResearchIteration] = []
    online_results: Dict = dict()
    code_results: Dict = dict()
    operator_results: List[OperatorRun] = []
    compiled_references: List[Any] = []
    inferred_queries: List[Any] = []
    attached_file_context = gather_raw_query_files(query_files)

    generated_images: List[str] = []
    generated_mermaidjs_diagram: str = None
    generated_asset_results: Dict = dict()
    program_execution_context: List[str] = []
    user_message_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    # Create a task to monitor for disconnections
    disconnect_monitor_task = None

    async def monitor_disconnection():
        nonlocal q, defiltered_query
        if isinstance(request_obj, Request):
            try:
                msg = await request_obj.receive()
                if msg["type"] == "http.disconnect":
                    logger.debug(f"Request cancelled. User {user} disconnected from {common.client} client.")
                    cancellation_event.set()
                    # ensure partial chat state saved on interrupt
                    # shield the save against task cancellation
                    if conversation:
                        await asyncio.shield(
                            save_to_conversation_log(
                                q,
                                chat_response="",
                                user=user,
                                compiled_references=compiled_references,
                                online_results=online_results,
                                code_results=code_results,
                                operator_results=operator_results,
                                research_results=research_results,
                                inferred_queries=inferred_queries,
                                client_application=user_scope.client_app,
                                conversation_id=conversation_id,
                                query_images=uploaded_images,
                                train_of_thought=train_of_thought,
                                raw_query_files=raw_query_files,
                                generated_images=generated_images,
                                generated_mermaidjs_diagram=generated_mermaidjs_diagram,
                                user_message_time=user_message_time,
                                tracer=tracer,
                            )
                        )
            except Exception as e:
                logger.error(f"Error in disconnect monitor: {e}")
        elif isinstance(request_obj, WebSocket):
            while request_obj.client_state == WebSocketState.CONNECTED and not cancellation_event.is_set():
                await asyncio.sleep(1)

                # Check if any interrupt query is received
                if interrupt_query := get_message_from_queue(parent_interrupt_queue):
                    if interrupt_query == ChatEvent.END_EVENT.value:
                        cancellation_event.set()
                        logger.debug(f"Chat cancelled by user {user} via interrupt queue.")
                    elif interrupt_query == ChatEvent.INTERRUPT.value:
                        cancellation_event.set()
                        logger.debug("Chat interrupted.")
                    else:
                        # Pass the interrupt query to child tasks
                        logger.info(f"Continuing chat with the new instruction: {interrupt_query}")
                        await child_interrupt_queue.put(interrupt_query)
                        # Append the interrupt query to the main query
                        q += f"\n\n{interrupt_query}"
                        defiltered_query += f"\n\n{defilter_query(interrupt_query)}"

            logger.debug(f"WebSocket disconnected or chat cancelled by user {user} from {common.client} client.")
            if conversation and cancellation_event.is_set():
                await asyncio.shield(
                    save_to_conversation_log(
                        q,
                        chat_response="",
                        user=user,
                        compiled_references=compiled_references,
                        online_results=online_results,
                        code_results=code_results,
                        operator_results=operator_results,
                        research_results=research_results,
                        inferred_queries=inferred_queries,
                        client_application=user_scope.client_app,
                        conversation_id=conversation_id,
                        query_images=uploaded_images,
                        train_of_thought=train_of_thought,
                        raw_query_files=raw_query_files,
                        generated_images=generated_images,
                        generated_mermaidjs_diagram=generated_mermaidjs_diagram,
                        user_message_time=user_message_time,
                        tracer=tracer,
                    )
                )

    # Cancel the disconnect monitor task if it is still running
    async def cancel_disconnect_monitor():
        if disconnect_monitor_task and not disconnect_monitor_task.done():
            logger.debug(f"Cancelling disconnect monitor task for user {user}")
            disconnect_monitor_task.cancel()
            try:
                await disconnect_monitor_task
            except asyncio.CancelledError:
                pass

    async def send_event(event_type: ChatEvent, data: str | dict):
        nonlocal ttft, train_of_thought
        if cancellation_event.is_set():
            return
        try:
            if event_type == ChatEvent.END_LLM_RESPONSE:
                collect_telemetry()
            elif event_type == ChatEvent.START_LLM_RESPONSE:
                ttft = time.perf_counter() - start_time
            elif event_type == ChatEvent.STATUS:
                train_of_thought.append({"type": event_type.value, "data": data})
            elif event_type == ChatEvent.THOUGHT:
                # Append the data to the last thought as thoughts are streamed
                if (
                    len(train_of_thought) > 0
                    and train_of_thought[-1]["type"] == ChatEvent.THOUGHT.value
                    and isinstance(train_of_thought[-1]["data"], str)
                    and isinstance(data, str)
                ):
                    train_of_thought[-1]["data"] += data
                else:
                    train_of_thought.append({"type": event_type.value, "data": data})

            if event_type == ChatEvent.MESSAGE:
                yield data
            elif event_type == ChatEvent.REFERENCES or ChatEvent.METADATA or stream:
                yield json.dumps({"type": event_type.value, "data": data}, ensure_ascii=False)
        except Exception as e:
            if not cancellation_event.is_set():
                logger.error(
                    f"Failed to stream chat API response to {user} on {common.client}: {e}.",
                    exc_info=True,
                )
        finally:
            if not cancellation_event.is_set():
                yield ChatEvent.END_EVENT.value
            # Cancel the disconnect monitor task if it is still running
            if cancellation_event.is_set() or event_type == ChatEvent.END_RESPONSE:
                await cancel_disconnect_monitor()

    async def send_llm_response(response: str, usage: dict = None):
        # Check if the client is still connected
        if cancellation_event.is_set():
            return
        # Send Chat Response
        async for result in send_event(ChatEvent.START_LLM_RESPONSE, ""):
            yield result
        async for result in send_event(ChatEvent.MESSAGE, response):
            yield result
        async for result in send_event(ChatEvent.END_LLM_RESPONSE, ""):
            yield result
        # Send Usage Metadata once llm interactions are complete
        if usage:
            async for event in send_event(ChatEvent.USAGE, usage):
                yield event
        async for result in send_event(ChatEvent.END_RESPONSE, ""):
            yield result

    def collect_telemetry():
        # Gather chat response telemetry
        nonlocal chat_metadata
        latency = time.perf_counter() - start_time
        cmd_set = set([cmd.value for cmd in conversation_commands])
        cost = (tracer.get("usage", {}) or {}).get("cost", 0)
        chat_metadata = chat_metadata or {}
        chat_metadata["conversation_command"] = cmd_set
        chat_metadata["agent"] = conversation.agent.slug if conversation and conversation.agent else None
        chat_metadata["cost"] = f"{cost:.5f}"
        chat_metadata["latency"] = f"{latency:.3f}"
        if ttft:
            chat_metadata["ttft_latency"] = f"{ttft:.3f}"
            logger.info(f"Chat response time to first token: {ttft:.3f} seconds")
        logger.info(f"Chat response total time: {latency:.3f} seconds")
        logger.info(f"Chat response cost: ${cost:.5f}")
        update_telemetry_state(
            request=request_obj,
            telemetry_type="api",
            api="chat",
            client=common.client,
            user_agent=headers.get("user-agent"),
            host=headers.get("host"),
            metadata=chat_metadata,
        )

    # Start the disconnect monitor in the background
    disconnect_monitor_task = asyncio.create_task(monitor_disconnection())

    if is_query_empty(q):
        async for result in send_llm_response("Please ask your query to get started.", tracer.get("usage")):
            yield result
        return

    # Automated tasks are handled before to allow mixing them with other conversation commands
    cmds_to_rate_limit = []
    if q.startswith("/automated_task"):
        q = q.replace("/automated_task", "").lstrip()
        cmds_to_rate_limit += [ConversationCommand.AutomatedTask]

    # Extract conversation command from query
    conversation_commands = [get_conversation_command(query=q)]

    conversation = await ConversationAdapters.aget_conversation_by_user(
        user,
        client_application=user_scope.client_app,
        conversation_id=conversation_id,
        title=title,
        create_new=body.create_new,
    )
    if not conversation:
        async for result in send_llm_response(f"Conversation {conversation_id} not found", tracer.get("usage")):
            yield result
        return
    conversation_id = str(conversation.id)

    async for event in send_event(ChatEvent.METADATA, {"conversationId": conversation_id, "turnId": turn_id}):
        yield event

    agent: Agent | None = None
    default_agent = await AgentAdapters.aget_default_agent()
    if conversation.agent and conversation.agent != default_agent:
        agent = conversation.agent

    if not conversation.agent:
        conversation.agent = default_agent
        await conversation.asave()
        agent = default_agent

    await is_ready_to_chat(user)
    user_name = await aget_user_name(user)
    location = None
    if city or region or country or country_code:
        location = LocationData(city=city, region=region, country=country, country_code=country_code)
    chat_history = conversation.messages

    # Get most recent memories and long term relevant memories if memory is enabled
    relevant_memories = []
    if await ConversationAdapters.ais_memory_enabled(user):
        recent_memories = await UserMemoryAdapters.pull_memories(user=user, agent=agent)
        long_term_memories = await UserMemoryAdapters.search_memories(query=q, user=user, agent=agent)
        # Create a de-duped set of memories
        relevant_memories = list({m.id: m for m in recent_memories + long_term_memories}.values())

    # If interrupted message in DB
    if last_message := await conversation.pop_message(interrupted=True):
        # Populate context from interrupted message
        online_results = {key: val.model_dump() for key, val in last_message.onlineContext.items() or []}
        code_results = {key: val.model_dump() for key, val in last_message.codeContext.items() or []}
        compiled_references = [ref.model_dump() for ref in last_message.context or []]
        research_results = [
            ResearchIteration(**iter_dict)
            for iter_dict in last_message.researchContext or []
            if iter_dict.get("summarizedResult")
        ]
        operator_results = [OperatorRun(**iter_dict) for iter_dict in last_message.operatorContext or []]
        train_of_thought = [thought.model_dump() for thought in last_message.trainOfThought or []]
        logger.info(f"Loaded interrupted partial context from conversation {conversation_id}.")

    if conversation_commands == [ConversationCommand.Default]:
        try:
            chosen_io = await aget_data_sources_and_output_format(
                q,
                chat_history,
                user=user,
                query_images=uploaded_images,
                agent=agent,
                query_files=attached_file_context,
                relevant_memories=relevant_memories,
                tracer=tracer,
            )
        except ValueError as e:
            logger.error(f"Error getting data sources and output format: {e}. Falling back to default.")
            chosen_io = {"sources": [ConversationCommand.General], "output": ConversationCommand.Text}

        conversation_commands = chosen_io.get("sources") + [chosen_io.get("output")]

        # If we're doing research, we don't want to do anything else
        if ConversationCommand.Research in conversation_commands:
            conversation_commands = [ConversationCommand.Research]

        conversation_commands_str = ", ".join([cmd.value for cmd in conversation_commands])
        async for result in send_event(ChatEvent.STATUS, f"**Selected Tools:** {conversation_commands_str}"):
            yield result

    cmds_to_rate_limit += conversation_commands
    for cmd in cmds_to_rate_limit:
        try:
            await conversation_command_rate_limiter.update_and_check_if_valid(request_obj, cmd)
            q = q.replace(f"/{cmd.value}", "").strip()
        except HTTPException as e:
            async for result in send_llm_response(str(e.detail), tracer.get("usage")):
                yield result
            return

    defiltered_query = defilter_query(q)

    if conversation_commands == [ConversationCommand.Research]:
        async for research_result in research(
            user=user,
            query=defiltered_query,
            conversation_id=conversation_id,
            conversation_history=chat_history,
            previous_iterations=list(research_results),
            query_images=uploaded_images,
            query_files=attached_file_context,
            relevant_memories=relevant_memories,
            user_name=user_name,
            location=location,
            send_status_func=partial(send_event, ChatEvent.STATUS),
            cancellation_event=cancellation_event,
            interrupt_queue=child_interrupt_queue,
            abort_message=ChatEvent.END_EVENT.value,
            agent=agent,
            tracer=tracer,
        ):
            if isinstance(research_result, ResearchIteration):
                if research_result.summarizedResult:
                    if research_result.onlineContext:
                        online_results.update(research_result.onlineContext)
                    if research_result.codeContext:
                        code_results.update(research_result.codeContext)
                    if research_result.context:
                        compiled_references.extend(research_result.context)
                if not research_results or research_results[-1] is not research_result:
                    research_results.append(research_result)
            else:
                yield research_result

            # Track operator results across research and operator iterations
            # This relies on two conditions:
            # 1. Check to append new (partial) operator results
            #    Relies on triggering this check on every status updates.
            #    Status updates cascade up from operator to research to chat api on every step.
            # 2. Keep operator results in sync with each research operator step
            #    Relies on python object references to ensure operator results
            #    are implicitly kept in sync after the initial append
            if (
                research_results
                and research_results[-1].operatorContext
                and (not operator_results or operator_results[-1] is not research_results[-1].operatorContext)
            ):
                operator_results.append(research_results[-1].operatorContext)

        # researched_results = await extract_relevant_info(q, researched_results, agent)
        if state.verbose > 1:
            logger.debug(f"Researched Results: {''.join(r.summarizedResult or '' for r in research_results)}")

    # Gather Context
    ## Gather Document References
    if ConversationCommand.Notes in conversation_commands:
        try:
            async for result in search_documents(
                q,
                (n or 7),
                d,
                user,
                chat_history,
                conversation_id,
                conversation_commands,
                location,
                partial(send_event, ChatEvent.STATUS),
                query_images=uploaded_images,
                query_files=attached_file_context,
                relevant_memories=relevant_memories,
                agent=agent,
                tracer=tracer,
            ):
                if isinstance(result, dict) and ChatEvent.STATUS in result:
                    yield result[ChatEvent.STATUS]
                else:
                    compiled_references.extend(result[0])
                    inferred_queries.extend(result[1])
                    defiltered_query = result[2]
        except Exception as e:
            error_message = f"Error searching knowledge base: {e}. Attempting to respond without document references."
            logger.error(error_message, exc_info=True)
            async for result in send_event(
                ChatEvent.STATUS, "Document search failed. I'll try respond without document references"
            ):
                yield result

        if not is_none_or_empty(compiled_references):
            distinct_headings = set([d.get("compiled").split("\n")[0] for d in compiled_references if "compiled" in d])
            distinct_files = set([d["file"] for d in compiled_references])
            # Strip only leading # from headings
            headings_str = "\n- " + "\n- ".join(distinct_headings).replace("#", "")
            async for result in send_event(
                ChatEvent.STATUS,
                f"**Found {len(distinct_headings)} Notes Across {len(distinct_files)} Files**: {headings_str}",
            ):
                yield result

        if conversation_commands == [ConversationCommand.Notes] and not await EntryAdapters.auser_has_entries(user):
            async for result in send_llm_response(f"{no_entries_found.format()}", tracer.get("usage")):
                yield result
            return

    if ConversationCommand.Notes in conversation_commands and is_none_or_empty(compiled_references):
        conversation_commands.remove(ConversationCommand.Notes)

    ## Gather Online References
    if ConversationCommand.Online in conversation_commands:
        try:
            async for result in search_online(
                defiltered_query,
                chat_history,
                location,
                user,
                partial(send_event, ChatEvent.STATUS),
                custom_filters=[],
                max_online_searches=3,
                query_images=uploaded_images,
                query_files=attached_file_context,
                relevant_memories=relevant_memories,
                agent=agent,
                tracer=tracer,
            ):
                if isinstance(result, dict) and ChatEvent.STATUS in result:
                    yield result[ChatEvent.STATUS]
                else:
                    online_results = result
        except Exception as e:
            error_message = f"Error searching online: {e}. Attempting to respond without online results"
            logger.warning(error_message)
            async for result in send_event(
                ChatEvent.STATUS, "Online search failed. I'll try respond without online references"
            ):
                yield result

    ## Gather Webpage References
    if ConversationCommand.Webpage in conversation_commands:
        try:
            async for result in read_webpages(
                defiltered_query,
                chat_history,
                location,
                user,
                partial(send_event, ChatEvent.STATUS),
                max_webpages_to_read=1,
                query_images=uploaded_images,
                query_files=attached_file_context,
                relevant_memories=relevant_memories,
                agent=agent,
                tracer=tracer,
            ):
                if isinstance(result, dict) and ChatEvent.STATUS in result:
                    yield result[ChatEvent.STATUS]
                else:
                    direct_web_pages = result
            webpages = []
            for query in direct_web_pages:
                if online_results.get(query):
                    online_results[query]["webpages"] = direct_web_pages[query]["webpages"]
                else:
                    online_results[query] = {"webpages": direct_web_pages[query]["webpages"]}

                for webpage in direct_web_pages[query]["webpages"]:
                    webpages.append(webpage["link"])
            async for result in send_event(ChatEvent.STATUS, f"**Read web pages**: {webpages}"):
                yield result
        except Exception as e:
            logger.warning(
                f"Error reading webpages: {e}. Attempting to respond without webpage results",
                exc_info=True,
            )
            async for result in send_event(
                ChatEvent.STATUS, "Webpage read failed. I'll try respond without webpage references"
            ):
                yield result

    ## Run Code
    if ConversationCommand.Code in conversation_commands:
        try:
            context = f"# Iteration 1:\n#---\nNotes:\n{compiled_references}\n\nOnline Results:{online_results}"
            async for result in run_code(
                defiltered_query,
                chat_history,
                context,
                location,
                user,
                partial(send_event, ChatEvent.STATUS),
                query_images=uploaded_images,
                query_files=attached_file_context,
                relevant_memories=relevant_memories,
                agent=agent,
                tracer=tracer,
            ):
                if isinstance(result, dict) and ChatEvent.STATUS in result:
                    yield result[ChatEvent.STATUS]
                else:
                    code_results = result
        except ValueError as e:
            program_execution_context.append("Failed to run code")
            logger.warning(
                f"Failed to use code tool: {e}. Attempting to respond without code results",
                exc_info=True,
            )

    ## Operate Computer
    if ConversationCommand.Operator in conversation_commands:
        try:
            async for result in operate_environment(
                defiltered_query,
                user,
                chat_history,
                location,
                list(operator_results)[-1] if operator_results else None,
                query_images=uploaded_images,
                query_files=attached_file_context,
                relevant_memories=relevant_memories,
                send_status_func=partial(send_event, ChatEvent.STATUS),
                agent=agent,
                cancellation_event=cancellation_event,
                interrupt_queue=child_interrupt_queue,
                tracer=tracer,
            ):
                if isinstance(result, dict) and ChatEvent.STATUS in result:
                    yield result[ChatEvent.STATUS]
                elif isinstance(result, OperatorRun):
                    if not operator_results or operator_results[-1] is not result:
                        operator_results.append(result)
                    # Add webpages visited while operating browser to references
                    if result.webpages:
                        if not online_results.get(defiltered_query):
                            online_results[defiltered_query] = {"webpages": result.webpages}
                        elif not online_results[defiltered_query].get("webpages"):
                            online_results[defiltered_query]["webpages"] = result.webpages
                        else:
                            online_results[defiltered_query]["webpages"] += result.webpages
        except ValueError as e:
            program_execution_context.append(f"Browser operation error: {e}")
            logger.warning(f"Failed to operate browser with {e}", exc_info=True)
            async for result in send_event(
                ChatEvent.STATUS, "Operating browser failed. I'll try respond appropriately"
            ):
                yield result

    ## Send Gathered References
    unique_online_results = deduplicate_organic_results(online_results)
    async for result in send_event(
        ChatEvent.REFERENCES,
        {
            "inferredQueries": inferred_queries,
            "context": compiled_references,
            "onlineContext": unique_online_results,
            "codeContext": code_results,
        },
    ):
        yield result

    # Generate Output
    ## Generate Image Output
    if ConversationCommand.Image in conversation_commands:
        async for result in text_to_image(
            defiltered_query,
            user,
            chat_history,
            location_data=location,
            references=compiled_references,
            online_results=online_results,
            send_status_func=partial(send_event, ChatEvent.STATUS),
            query_images=uploaded_images,
            query_files=attached_file_context,
            relevant_memories=relevant_memories,
            agent=agent,
            tracer=tracer,
        ):
            if isinstance(result, dict) and ChatEvent.STATUS in result:
                yield result[ChatEvent.STATUS]
            else:
                generated_image, status_code, improved_image_prompt = result

        inferred_queries.append(improved_image_prompt)
        if generated_image is None or status_code != 200:
            program_execution_context.append(f"Failed to generate image with {improved_image_prompt}")
            async for result in send_event(ChatEvent.STATUS, "Failed to generate image"):
                yield result
        else:
            generated_images.append(generated_image)

            generated_asset_results["images"] = {
                "description": improved_image_prompt,
            }

            async for result in send_event(
                ChatEvent.GENERATED_ASSETS,
                {
                    "images": [generated_image],
                },
            ):
                yield result

    if ConversationCommand.Diagram in conversation_commands:
        async for result in send_event(ChatEvent.STATUS, "Creating diagram"):
            yield result

        inferred_queries = []
        async for result in generate_mermaidjs_diagram(
            q=defiltered_query,
            chat_history=chat_history,
            location_data=location,
            note_references=compiled_references,
            online_results=online_results,
            query_images=uploaded_images,
            query_files=attached_file_context,
            relevant_memories=relevant_memories,
            user=user,
            agent=agent,
            send_status_func=partial(send_event, ChatEvent.STATUS),
            tracer=tracer,
        ):
            if isinstance(result, dict) and ChatEvent.STATUS in result:
                yield result[ChatEvent.STATUS]
            else:
                better_diagram_description_prompt, mermaidjs_diagram_description = result
                if better_diagram_description_prompt and mermaidjs_diagram_description:
                    inferred_queries.append(better_diagram_description_prompt)
                    generated_mermaidjs_diagram = mermaidjs_diagram_description

                    generated_asset_results["diagrams"] = {
                        "query": better_diagram_description_prompt,
                    }

                    async for result in send_event(
                        ChatEvent.GENERATED_ASSETS,
                        {
                            "mermaidjsDiagram": mermaidjs_diagram_description,
                        },
                    ):
                        yield result
                else:
                    error_message = "Failed to generate diagram. Please try again later."
                    program_execution_context.append(
                        prompts.failed_diagram_generation.format(attempted_diagram=better_diagram_description_prompt)
                    )

                    async for result in send_event(ChatEvent.STATUS, error_message):
                        yield result

    # Check if the user has disconnected
    if cancellation_event.is_set():
        logger.debug(f"Stopping LLM response to user {user} on {common.client} client.")
        # Cancel the disconnect monitor task if it is still running
        await cancel_disconnect_monitor()
        return

    ## Generate Text Output
    async for result in send_event(ChatEvent.STATUS, "**Generating a well-informed response**"):
        yield result

    llm_response, chat_metadata = await agenerate_chat_response(
        defiltered_query,
        chat_history,
        conversation,
        compiled_references,
        online_results,
        code_results,
        operator_results,
        research_results,
        user,
        location,
        user_name,
        uploaded_images,
        attached_file_context,
        relevant_memories,
        program_execution_context,
        generated_asset_results,
        is_subscribed,
        tracer,
    )

    full_response = ""
    message_start = True
    async for item in llm_response:
        # Should not happen with async generator. Skip.
        if item is None or not isinstance(item, ResponseWithThought):
            logger.warning(f"Unexpected item type in LLM response: {type(item)}. Skipping.")
            continue
        if cancellation_event.is_set():
            break
        message = item.text
        full_response += message if message else ""
        if item.thought:
            async for result in send_event(ChatEvent.THOUGHT, item.thought):
                yield result
            continue
        # Start sending response
        elif message_start:
            message_start = False
            async for result in send_event(ChatEvent.START_LLM_RESPONSE, ""):
                yield result

        try:
            async for result in send_event(ChatEvent.MESSAGE, message):
                yield result
        except Exception as e:
            if not cancellation_event.is_set():
                logger.warning(f"Error during streaming. Stopping send: {e}")
            break

    # Check if the user has disconnected
    if cancellation_event.is_set():
        logger.debug(f"Stopping LLM response to user {user} on {common.client} client.")
        # Cancel the disconnect monitor task if it is still running
        await cancel_disconnect_monitor()
        return

    # Save conversation once finish streaming
    asyncio.create_task(
        save_to_conversation_log(
            q,
            chat_response=full_response,
            user=user,
            compiled_references=compiled_references,
            online_results=online_results,
            code_results=code_results,
            operator_results=operator_results,
            research_results=research_results,
            inferred_queries=inferred_queries,
            client_application=user_scope.client_app,
            conversation_id=str(conversation.id),
            query_images=uploaded_images,
            train_of_thought=train_of_thought,
            raw_query_files=raw_query_files,
            relevant_memories=relevant_memories,
            generated_images=generated_images,
            generated_mermaidjs_diagram=generated_mermaidjs_diagram,
            tracer=tracer,
        )
    )

    # Signal end of LLM response after the loop finishes
    async for result in send_event(ChatEvent.END_LLM_RESPONSE, ""):
        yield result

    # Send Usage Metadata once llm interactions are complete
    if tracer.get("usage"):
        async for event in send_event(ChatEvent.USAGE, tracer.get("usage")):
            yield event
    async for result in send_event(ChatEvent.END_RESPONSE, ""):
        yield result
    logger.debug("Finished streaming response")

    # Cancel the disconnect monitor task if it is still running
    await cancel_disconnect_monitor()


@api_chat.websocket("/ws")
@requires(["authenticated"])
async def chat_ws(
    websocket: WebSocket,
    common: CommonQueryParams,
):
    # Validate WebSocket Origin
    origin = websocket.headers.get("origin")
    if not origin or URL(origin).hostname not in ALLOWED_HOSTS:
        await websocket.close(code=1008, reason="Origin not allowed")
        return

    # Limit open websocket connections per user
    user = websocket.scope["user"].object
    connection_manager = WebSocketConnectionManager(trial_user_max_connections=5, subscribed_user_max_connections=10)
    connection_id = str(uuid.uuid4())

    if not await connection_manager.can_connect(websocket):
        await websocket.close(code=1008, reason="Connection limit exceeded")
        logger.info(f"WebSocket connection rejected for user {user.id}: connection limit exceeded")
        return

    await websocket.accept()

    # Note new websocket connection for the user
    await connection_manager.register_connection(user, connection_id)

    # Initialize rate limiters
    rate_limiter_per_minute = ApiUserRateLimiter(requests=20, subscribed_requests=20, window=60, slug="chat_minute")
    rate_limiter_per_day = ApiUserRateLimiter(
        requests=100, subscribed_requests=600, window=60 * 60 * 24, slug="chat_day"
    )
    image_rate_limiter = ApiImageRateLimiter(max_images=10, max_combined_size_mb=20)

    # Shared interrupt queue for communicating interrupts to ongoing research
    interrupt_queue: asyncio.Queue = asyncio.Queue(maxsize=10)
    current_task = None

    try:
        while True:
            data = await websocket.receive_json()

            # Check if this is an interrupt message
            if data.get("type") == "interrupt":
                if current_task and not current_task.done():
                    # Send interrupt signal to the ongoing task
                    await interrupt_queue.put(data.get("query") or ChatEvent.END_EVENT.value)
                    logger.info(
                        f"Interrupt signal sent to ongoing task for user {websocket.scope['user'].object.id} with query: {data.get('query')}"
                    )
                    if data.get("query"):
                        ack_type = "interrupt_message_acknowledged"
                        await websocket.send_text(json.dumps({"type": ack_type}))
                    else:
                        ack_type = "interrupt_acknowledged"
                        await websocket.send_text(json.dumps({"type": ack_type}))
                else:
                    ack_type = "interrupt_acknowledged"
                    await websocket.send_text(json.dumps({"type": ack_type}))
                    logger.info(f"No ongoing task to interrupt for user {websocket.scope['user'].object.id}")
                continue

            # Handle regular chat messages - ensure data has required fields
            if "q" not in data:
                await websocket.send_text(json.dumps({"error": "Missing required field 'q' in chat message"}))
                continue

            body = ChatRequestBody(**data)

            # Apply rate limiting manually
            try:
                await rate_limiter_per_minute.check_websocket(websocket)
                await rate_limiter_per_day.check_websocket(websocket)
                image_rate_limiter.check_websocket(websocket, body)
            except HTTPException as e:
                await websocket.send_text(json.dumps({"error": e.detail}))
                continue

            # Cancel any ongoing task before starting a new one
            if current_task and not current_task.done():
                current_task.cancel()
                try:
                    await current_task
                except asyncio.CancelledError:
                    pass

            # Create a new task for processing the chat request
            current_task = asyncio.create_task(process_chat_request(websocket, body, common, interrupt_queue))

    except WebSocketDisconnect:
        logger.info(f"WebSocket disconnected for user {websocket.scope['user'].object.id}")
        if current_task and not current_task.done():
            interrupt_queue.put_nowait(ChatEvent.INTERRUPT.value)
    except Exception as e:
        logger.error(f"Error in websocket chat: {e}", exc_info=True)
        if current_task and not current_task.done():
            current_task.cancel()
        await websocket.close(code=1011, reason="Internal Server Error")
    finally:
        # Always unregister the connection on disconnect
        await connection_manager.unregister_connection(user, connection_id)


async def process_chat_request(
    websocket: WebSocket,
    body: ChatRequestBody,
    common: CommonQueryParams,
    interrupt_queue: asyncio.Queue,
):
    """Process a single chat request with interrupt support"""

    # Server-side message buffering for better streaming performance
    @dataclass
    class MessageBuffer:
        """Buffer for managing streamed chat messages with timing control."""

        content: str = ""
        timeout: Optional[asyncio.Task] = None
        last_flush: float = 0.0

        def __post_init__(self):
            """Initialize last_flush with current time if not provided."""
            if self.last_flush == 0.0:
                self.last_flush = time.perf_counter()

    message_buffer = MessageBuffer()
    thought_buffer = MessageBuffer()
    BUFFER_FLUSH_INTERVAL = 0.1  # 100ms buffer interval
    BUFFER_MAX_SIZE = 512  # Flush if buffer reaches this size

    async def flush_message_buffer():
        """Flush the accumulated message buffer to the client"""
        nonlocal message_buffer
        if message_buffer.content:
            buffered_content = message_buffer.content
            message_buffer.content = ""
            message_buffer.last_flush = time.perf_counter()
            if message_buffer.timeout:
                message_buffer.timeout.cancel()
                message_buffer.timeout = None
            yield buffered_content

    async def flush_thought_buffer():
        """Flush the accumulated thought buffer to the client"""
        nonlocal thought_buffer
        if thought_buffer.content:
            thought_event = json.dumps({"type": ChatEvent.THOUGHT.value, "data": thought_buffer.content})
            thought_buffer.content = ""
            thought_buffer.last_flush = time.perf_counter()
            if thought_buffer.timeout:
                thought_buffer.timeout.cancel()
                thought_buffer.timeout = None
            yield thought_event

    try:
        # Since we are using websockets, we can ignore the stream parameter and always stream
        response_iterator = event_generator(
            body,
            websocket.scope["user"],
            common,
            websocket.headers,
            websocket,
            interrupt_queue,
        )
        async for event in response_iterator:
            if not event:
                continue
            elif event.startswith("{") and event.endswith("}"):
                evt_json = json.loads(event)
                if evt_json["type"] == ChatEvent.END_LLM_RESPONSE.value:
                    # Flush remaining buffer content on end llm response event
                    chunks = "".join([chunk async for chunk in flush_message_buffer()])
                    await websocket.send_text(chunks)
                    await websocket.send_text(ChatEvent.END_EVENT.value)
                elif evt_json["type"] == ChatEvent.THOUGHT.value:
                    # Buffer THOUGHT events for better streaming performance
                    thought_buffer.content += str(evt_json.get("data", ""))

                    # Flush if buffer is too large or enough time has passed
                    current_time = time.perf_counter()
                    should_flush_time = (current_time - thought_buffer.last_flush) >= BUFFER_FLUSH_INTERVAL
                    should_flush_size = len(thought_buffer.content) >= BUFFER_MAX_SIZE

                    if should_flush_size or should_flush_time:
                        thought_event = "".join([chunk async for chunk in flush_thought_buffer()])
                        await websocket.send_text(thought_event)
                        await websocket.send_text(ChatEvent.END_EVENT.value)
                    else:
                        # Cancel any previous timeout tasks to reset the flush timer
                        if thought_buffer.timeout:
                            thought_buffer.timeout.cancel()

                        async def delayed_thought_flush():
                            """Flush thought buffer if no new messages arrive within debounce interval."""
                            await asyncio.sleep(BUFFER_FLUSH_INTERVAL)
                            # Check if there's still content to flush
                            thought_chunks = "".join([chunk async for chunk in flush_thought_buffer()])
                            if thought_chunks:
                                thought_event = "".join([chunk async for chunk in flush_thought_buffer()])
                                await websocket.send_text(thought_event)
                                await websocket.send_text(ChatEvent.END_EVENT.value)

                        # Flush buffer if no new thoughts arrive within debounce interval
                        thought_buffer.timeout = asyncio.create_task(delayed_thought_flush())
                    continue
                await websocket.send_text(event)
                await websocket.send_text(ChatEvent.END_EVENT.value)
            elif event != ChatEvent.END_EVENT.value:
                # Buffer MESSAGE events for better streaming performance
                message_buffer.content += str(event)

                # Flush if buffer is too large or enough time has passed
                current_time = time.perf_counter()
                should_flush_time = (current_time - message_buffer.last_flush) >= BUFFER_FLUSH_INTERVAL
                should_flush_size = len(message_buffer.content) >= BUFFER_MAX_SIZE

                if should_flush_size or should_flush_time:
                    chunks = "".join([chunk async for chunk in flush_message_buffer()])
                    await websocket.send_text(chunks)
                    await websocket.send_text(ChatEvent.END_EVENT.value)
                else:
                    # Cancel any previous timeout tasks to reset the flush timer
                    if message_buffer.timeout:
                        message_buffer.timeout.cancel()

                    async def delayed_flush():
                        """Flush message buffer if no new messages arrive within debounce interval."""
                        await asyncio.sleep(BUFFER_FLUSH_INTERVAL)
                        # Check if there's still content to flush
                        chunks = "".join([chunk async for chunk in flush_message_buffer()])
                        await websocket.send_text(chunks)
                        await websocket.send_text(ChatEvent.END_EVENT.value)

                    # Flush buffer if no new messages arrive within debounce interval
                    message_buffer.timeout = asyncio.create_task(delayed_flush())
    except asyncio.CancelledError:
        logger.debug(f"Chat request cancelled for user {websocket.scope['user'].object.id}")
        raise
    except Exception as e:
        await websocket.send_text(json.dumps({"error": "Internal server error"}))
        logger.error(f"Error processing chat request: {e}", exc_info=True)
        raise


@api_chat.post("")
@requires(["authenticated"])
async def chat(
    request: Request,
    common: CommonQueryParams,
    body: ChatRequestBody,
    rate_limiter_per_minute=Depends(
        ApiUserRateLimiter(requests=20, subscribed_requests=20, window=60, slug="chat_minute")
    ),
    rate_limiter_per_day=Depends(
        ApiUserRateLimiter(requests=100, subscribed_requests=600, window=60 * 60 * 24, slug="chat_day")
    ),
    image_rate_limiter=Depends(ApiImageRateLimiter(max_images=10, max_combined_size_mb=20)),
):
    response_iterator = event_generator(
        body,
        request.user,
        common,
        request.headers,
        request,
    )

    # Stream Text Response
    if body.stream:
        return StreamingResponse(response_iterator, media_type="text/plain")
    # Non-Streaming Text Response
    else:
        response_data = await read_chat_stream(response_iterator)
        return Response(content=json.dumps(response_data), media_type="application/json", status_code=200)
